"""add metadata column for datasets

Revision ID: c1a0b0ee712e
Revises: 5a5105707147
Create Date: 2022-05-20 14:37:59.155025

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "c1a0b0ee712e"
down_revision = "5a5105707147"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("base_dataset", sa.Column("metadata", sa.JSON(), nullable=True))

    op.execute(sa.text("UPDATE base_dataset SET metadata = extra::jsonb->'metadata'"))
    op.execute(sa.text("UPDATE base_dataset SET extra = (extra::jsonb - 'metadata')"))
    op.execute(sa.text("UPDATE base_dataset SET metadata = '{}'::json WHERE metadata IS NULL"))

    op.alter_column("base_dataset", "metadata", nullable=False)

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""make compute tag not nullable

Revision ID: 4485daecc21b
Revises: 435c1f35227d
Create Date: 2022-02-09 13:19:12.191737

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4485daecc21b"
down_revision = "435c1f35227d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.execute(sa.text("UPDATE service_queue SET tag = '*' WHERE tag IS NULL"))
    op.execute(sa.text("UPDATE task_queue SET tag = '*' WHERE tag IS NULL"))

    op.alter_column("service_queue", "tag", existing_type=sa.VARCHAR(), nullable=False)
    op.alter_column("task_queue", "tag", existing_type=sa.VARCHAR(), nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("task_queue", "tag", existing_type=sa.VARCHAR(), nullable=True)
    op.alter_column("service_queue", "tag", existing_type=sa.VARCHAR(), nullable=True)
    # ### end Alembic commands ###



"""update users and roles

Revision ID: ace6189b4a68
Revises: 038ffd952a00
Create Date: 2021-03-13 15:53:00.374469

"""

import logging

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from migration_helpers.default_roles import default_roles

# revision identifiers, used by Alembic.
revision = "ace6189b4a68"
down_revision = "038ffd952a00"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "role",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("rolename", sa.String(), nullable=False),
        sa.Column("permissions", sa.JSON(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("rolename", name="ux_role_rolename"),
    )

    # Create some temporary tables for updating
    user_table = table(
        "user",
        column("username", sa.String),
        column("permissions", postgresql.JSON(astext_type=sa.Text)),
        column("role_id", sa.Integer),
    )

    role_table = table(
        "role",
        column("id", sa.Integer),
        column("rolename", sa.String),
        column("permissions", sa.JSON),
    )

    bind = op.get_bind()
    session = Session(bind=bind)

    # Populate the role table with with the default roles
    for rolename, permissions in default_roles.items():
        session.execute(role_table.insert().values({"rolename": rolename, "permissions": permissions}))
    session.flush()

    # Now get the ids of the just-added roles
    role_query = session.query(role_table).all()
    role_map = {r.rolename: r.id for r in role_query}

    # Now modify the user table
    op.add_column(
        "user",
        sa.Column("enabled", sa.Boolean(), nullable=False, server_default="True"),
    )
    op.add_column("user", sa.Column("email", sa.String(), nullable=False, server_default=""))
    op.add_column("user", sa.Column("fullname", sa.String(), nullable=False, server_default=""))
    op.add_column(
        "user",
        sa.Column("organization", sa.String(), nullable=False, server_default=""),
    )

    # Set as nullable first, then change after it is populated
    op.add_column("user", sa.Column("role_id", sa.Integer(), nullable=True))

    op.drop_constraint("user_username_key", "user", type_="unique")
    op.create_unique_constraint("ux_user_username", "user", ["username"])
    op.create_foreign_key(None, "user", "role", ["role_id"], ["id"])

    # For all users, determine what the new role should be
    users = session.query(user_table).all()
    for user in users:
        permissions = set(user.permissions)
        if "admin" in permissions:
            role_id = role_map["admin"]
        elif permissions == {"read"}:
            role_id = role_map["read"]
        elif permissions <= {"write", "compute"}:
            role_id = role_map["submit"]
        elif permissions == {"queue"}:
            role_id = role_map["compute"]
        else:
            logging.getLogger().warning(
                f"!!!! User {user.username} does not have permissions that map to a single role. YOU MUST MANUALLY FIX THIS. Setting to 'read' for now"
            )
            role_id = role_map["read"]

        session.execute(user_table.update().where(user_table.c.username == user.username).values({"role_id": role_id}))

    session.flush()

    # Now we can drop the old permissions column and make the role_id column not nullable
    op.drop_column("user", "permissions")
    op.alter_column("user", "role_id", existing_type=sa.Integer(), nullable=False)

    session.commit()
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "user",
        sa.Column(
            "permissions",
            postgresql.JSON(astext_type=sa.Text()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_constraint(None, "user", type_="foreignkey")
    op.drop_index(op.f("ix_user_username"), table_name="user")
    op.create_unique_constraint("user_username_key", "user", ["username"])
    op.drop_column("user", "role_id")
    op.drop_column("user", "organization")
    op.drop_column("user", "fullname")
    op.drop_column("user", "enabled")
    op.drop_column("user", "email")
    op.drop_table("role")

    logging.getLogger().warning(f"!!!! All user permissions have been set to null. You must change this manually")

    # ### end Alembic commands ###



"""add missing indices

Revision ID: f3ad208b70da
Revises: 37bf1ec2b1d1
Create Date: 2022-04-07 15:45:52.735478

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "f3ad208b70da"
down_revision = "37bf1ec2b1d1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index("ix_record_comment_record_id", "record_comment", ["record_id"], unique=False)
    op.create_index("ix_record_info_backup_record_id", "record_info_backup", ["record_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_record_info_backup_record_id", table_name="record_info_backup")
    op.drop_index("ix_record_comment_record_id", table_name="record_comment")
    # ### end Alembic commands ###



"""clean up native files

Revision ID: 1a61b3bb1ee4
Revises: 1c46a35bf565
Create Date: 2023-02-08 12:22:48.401847

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1a61b3bb1ee4"
down_revision = "1c46a35bf565"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("native_file", "is_text")
    op.drop_column("native_file", "uncompressed_size")

    op.alter_column("native_file", "compression", new_column_name="compression_type")

    op.execute(sa.text("ALTER TABLE native_file ALTER COLUMN data SET STORAGE EXTERNAL"))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")



"""collection type not nullable

Revision ID: b74cd7a31b69
Revises: ace6189b4a68
Create Date: 2021-04-21 15:40:24.263298

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b74cd7a31b69"
down_revision = "ace6189b4a68"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # All collection_types that were null should now be 'collection'
    # until this is changed later
    op.execute("UPDATE collection SET collection_type='collection' WHERE collection_type is null")

    op.alter_column("collection", "collection_type", existing_type=sa.VARCHAR(), nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("collection", "collection_type", existing_type=sa.VARCHAR(), nullable=True)

    op.execute("UPDATE collection SET collection_type=null WHERE collection_type = 'collection'")
    # ### end Alembic commands ###



"""add largebinary storage

Revision ID: 4122a71f7031
Revises: ca9c57c0df8d
Create Date: 2023-01-28 09:29:04.763067

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import ENUM

# revision identifiers, used by Alembic.
revision = "4122a71f7031"
down_revision = "ca9c57c0df8d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    compressionenum = ENUM(name="compressionenum", create_type=False)
    op.create_table(
        "largebinary_store",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.Column("size", sa.BigInteger(), nullable=False),
        sa.Column("checksum", sa.String(), nullable=False),
        sa.Column("compression_type", compressionenum, nullable=False),
        sa.Column("data_local", sa.LargeBinary(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_foreign_key(None, "largebinary_store", "base_record", ["record_id"], ["id"], ondelete="cascade")

    op.execute(sa.text("ALTER TABLE largebinary_store ALTER COLUMN data_local SET STORAGE EXTERNAL"))

    # Now migrate task queue
    # We can just drop the function kwargs column. The data will be re-generated if we also delete the
    # data in the 'function' column
    op.add_column("task_queue", sa.Column("function_kwargs_lb_id", sa.Integer(), nullable=True))
    op.create_index("ix_task_queue_function_kwargs_lb_id", "task_queue", ["function_kwargs_lb_id"], unique=False)
    op.create_foreign_key(None, "task_queue", "largebinary_store", ["function_kwargs_lb_id"], ["id"])
    op.drop_column("task_queue", "function_kwargs")
    op.execute(sa.text("UPDATE task_queue SET function = NULL"))

    # Service subtasks will be empty for this migration, so don't migrate any data
    op.add_column("service_subtask_record", sa.Column("function_kwargs_lb_id", sa.Integer(), nullable=True))
    op.add_column("service_subtask_record", sa.Column("results_lb_id", sa.Integer(), nullable=True))
    op.create_index(
        "ix_service_subtask_record_function_kwargs_lb_id",
        "service_subtask_record",
        ["function_kwargs_lb_id"],
        unique=False,
    )
    op.create_index(
        "ix_service_subtask_record_results_lb_id",
        "service_subtask_record",
        ["results_lb_id"],
        unique=False,
    )
    op.create_foreign_key(None, "service_subtask_record", "largebinary_store", ["function_kwargs_lb_id"], ["id"])
    op.create_foreign_key(None, "service_subtask_record", "largebinary_store", ["results_lb_id"], ["id"])
    op.drop_column("service_subtask_record", "function_kwargs")
    op.drop_column("service_subtask_record", "results")
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""remove largebinary

Revision ID: 698131e6ffd5
Revises: 363c647d7840
Create Date: 2023-05-07 11:24:42.527020

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "698131e6ffd5"
down_revision = "363c647d7840"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # We assume no service subtask records for this migration
    conn = op.get_bind()
    r = conn.execute(sa.text("SELECT COUNT(*) FROM service_subtask_record"))
    n_subtasks = r.scalar_one()

    if n_subtasks > 0:
        raise RuntimeError(f"Cannot upgrade - there are {n_subtasks} existing service subtask records")

    # New columns in service_subtask_record are just plain JSON
    op.add_column("service_subtask_record", sa.Column("function_kwargs", sa.JSON(), nullable=False))
    op.add_column("service_subtask_record", sa.Column("results", sa.JSON(), nullable=True))

    # Remove old columns/indexes of service_subtask_record
    op.drop_index("ix_service_subtask_record_function_kwargs_lb_id", table_name="service_subtask_record")
    op.drop_index("ix_service_subtask_record_results_lb_id", table_name="service_subtask_record")
    op.drop_constraint(
        "service_subtask_record_function_kwargs_lb_id_fkey", "service_subtask_record", type_="foreignkey"
    )
    op.drop_constraint("service_subtask_record_results_lb_id_fkey", "service_subtask_record", type_="foreignkey")
    op.drop_column("service_subtask_record", "results_lb_id")
    op.drop_column("service_subtask_record", "function_kwargs_lb_id")

    # Now the task queue
    op.add_column("task_queue", sa.Column("function_kwargs_compressed", sa.LargeBinary(), nullable=True))
    op.drop_index("ix_task_queue_function_kwargs_lb_id", table_name="task_queue")
    op.drop_constraint("task_queue_function_kwargs_lb_id_fkey", "task_queue", type_="foreignkey")
    op.drop_column("task_queue", "function_kwargs_lb_id")

    # Remove the function from existing tasks. They will be autogenerated
    op.execute(sa.text("UPDATE task_queue SET function = NULL"))

    op.drop_table("largebinary_store")

    # Drop various functions and triggers
    op.execute(sa.text('drop trigger if exists "qca_service_subtask_delete_lb_tr" on "service_subtask_record";'))
    op.execute(sa.text('drop trigger if exists "qca_task_queue_delete_lb_tr" on "task_queue";'))
    op.execute(sa.text('drop function if exists "qca_largebinary_base_delete";'))
    op.execute(sa.text('drop function if exists "qca_service_subtask_delete_lb";'))
    op.execute(sa.text('drop function if exists "qca_task_queue_delete_lb";'))

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_queue", sa.Column("function_kwargs_lb_id", sa.INTEGER(), autoincrement=False, nullable=True))
    op.create_foreign_key(
        "task_queue_function_kwargs_lb_id_fkey", "task_queue", "largebinary_store", ["function_kwargs_lb_id"], ["id"]
    )
    op.create_index("ix_task_queue_function_kwargs_lb_id", "task_queue", ["function_kwargs_lb_id"], unique=False)
    op.drop_column("task_queue", "function_kwargs_compressed")
    op.add_column(
        "service_subtask_record", sa.Column("function_kwargs_lb_id", sa.INTEGER(), autoincrement=False, nullable=True)
    )
    op.add_column(
        "service_subtask_record", sa.Column("results_lb_id", sa.INTEGER(), autoincrement=False, nullable=True)
    )
    op.create_foreign_key(
        "service_subtask_record_results_lb_id_fkey",
        "service_subtask_record",
        "largebinary_store",
        ["results_lb_id"],
        ["id"],
    )
    op.create_foreign_key(
        "service_subtask_record_function_kwargs_lb_id_fkey",
        "service_subtask_record",
        "largebinary_store",
        ["function_kwargs_lb_id"],
        ["id"],
    )
    op.create_index(
        "ix_service_subtask_record_results_lb_id", "service_subtask_record", ["results_lb_id"], unique=False
    )
    op.create_index(
        "ix_service_subtask_record_function_kwargs_lb_id",
        "service_subtask_record",
        ["function_kwargs_lb_id"],
        unique=False,
    )
    op.drop_column("service_subtask_record", "results")
    op.drop_column("service_subtask_record", "function_kwargs")
    op.create_table(
        "largebinary_store",
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column("size", sa.BIGINT(), autoincrement=False, nullable=False),
        sa.Column("checksum", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column(
            "compression_type",
            postgresql.ENUM("none", "lzma", "zstd", name="compressionenum"),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("data_local", postgresql.BYTEA(), autoincrement=False, nullable=False),
        sa.Column("largebinary_type", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.PrimaryKeyConstraint("id", name="largebinary_store_pkey"),
    )
    # ### end Alembic commands ###



"""migrate_status_and_tasks

Revision ID: 98aea37d208d
Revises: 4257f1814d0d
Create Date: 2021-06-25 09:34:07.378008

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import func, column

# revision identifiers, used by Alembic.
revision = "98aea37d208d"
down_revision = "4257f1814d0d"
branch_labels = None
depends_on = None


def upgrade():
    # Queue manager
    op.drop_column("queue_manager", "procedures")

    # We are going to cheat to make this faster
    # First, we remove all 'incomplete' by replacing with the task/service status
    # Then, we rename 'incomplete' to 'invalid', which is a new value.
    # I tried doing creating a new type and then doing an in-place ALTER COLUMN,
    # but took way too long and rewrote the whole table (?)

    # Migrate the status in base result to reflect any status in the task/service queues
    # The only thing to really change is 'incomplete' base_result rows. Change those to be the status
    # of the task/service)
    op.execute(
        sa.text(
            """UPDATE base_result
           SET status = task_queue.status::text::recordstatusenum
           FROM task_queue
           WHERE base_result.id = task_queue.base_result_id
           AND base_result.status = 'incomplete'"""
        )
    )

    op.execute(
        sa.text(
            """UPDATE base_result
           SET status = service_queue.status::text::recordstatusenum
           FROM service_queue
           WHERE base_result.id = service_queue.procedure_id
           AND base_result.status = 'incomplete'"""
        )
    )

    # Delete all tasks/services for completed records
    op.execute(
        sa.text(
            """
               DELETE FROM task_queue
               USING base_result
               WHERE base_result.id = task_queue.base_result_id
               AND base_result.status = 'complete'
               """
        )
    )

    # Delete all tasks/services for completed records
    op.execute(
        sa.text(
            """
               DELETE FROM service_queue
               USING base_result
               WHERE base_result.id = service_queue.procedure_id
               AND base_result.status = 'complete'
               """
        )
    )

    # Also mark all results without a corresponding task or service as "cancelled"
    op.execute(
        sa.text(
            """
               UPDATE base_result
               SET status = 'cancelled'
               WHERE id IN (
                   SELECT base_result.id FROM base_result
                   LEFT OUTER JOIN task_queue ON task_queue.base_result_id = base_result.id
                   WHERE base_result.status != 'complete'
                   AND base_result.result_type IN ('result', 'optimization_procedure')
                   AND task_queue.base_result_id IS NULL
               )
               """
        )
    )

    op.execute(
        sa.text(
            """
               UPDATE base_result
               SET status = 'cancelled'
               WHERE id IN (
                   SELECT base_result.id FROM base_result
                   LEFT OUTER JOIN service_queue ON service_queue.procedure_id = base_result.id
                   WHERE base_result.status != 'complete'
                   AND base_result.result_type IN ('torsiondrive_procedure', 'gridoptimization_procedure')
                   AND service_queue.procedure_id IS NULL
               )
               """
        )
    )

    # Reset tasks that are assigned to inactive managers
    op.execute(
        sa.text(
            """
               UPDATE base_result
               SET status = 'waiting'
               WHERE id IN (
                   SELECT base_result.id FROM base_result
                   LEFT OUTER JOIN queue_manager ON base_result.manager_name = queue_manager.name
                   WHERE base_result.status = 'running'
                   AND base_result.result_type IN ('result', 'optimization_procedure')
                   AND queue_manager.status = 'inactive'
               )
               """
        )
    )

    # There should be no more incompletes. Check
    session = Session(op.get_bind())
    r = session.execute(sa.text("""SELECT id FROM base_result WHERE status = 'incomplete'""")).scalars().first()
    assert r is None

    # Now rename 'incomplete' to 'invalid'
    op.execute(sa.text("ALTER TYPE recordstatusenum RENAME VALUE 'incomplete' TO 'invalid'"))

    # Now do the required_programs column of the task_queue
    # Form this from the program and procedure columns
    op.add_column("task_queue", sa.Column("required_programs", postgresql.ARRAY(sa.TEXT()), nullable=True))
    op.execute(
        sa.text(
            "UPDATE task_queue SET required_programs = ARRAY[LOWER(program), LOWER(procedure)] WHERE procedure IS NOT NULL"
        )
    )
    op.execute(sa.text("UPDATE task_queue SET required_programs = ARRAY[LOWER(program)] WHERE procedure IS NULL"))
    op.alter_column("task_queue", "required_programs", existing_type=postgresql.JSONB, nullable=False)
    op.create_check_constraint(
        "ck_task_queue_requirements_lower",
        "task_queue",
        column("required_programs").cast(sa.TEXT) == func.lower(column("required_programs").cast(sa.TEXT)),
    )

    # OTHER MANIPULATIONS BELOW
    # Task Queue
    # NOTE: We are leaving the manager column in place for a future migration
    op.drop_index("ix_task_queue_created_on", table_name="task_queue")
    op.drop_index("ix_task_queue_keys", table_name="task_queue")
    op.drop_index("ix_task_waiting_sort", table_name="task_queue")

    op.create_index("ix_task_queue_required_programs", "task_queue", ["required_programs"], unique=False)
    op.create_index("ix_task_queue_tag", "task_queue", ["tag"], unique=False)
    op.execute("CREATE INDEX ix_task_queue_waiting_sort ON task_queue (priority desc, created_on)")

    op.drop_column("task_queue", "modified_on")
    op.drop_column("task_queue", "procedure")
    op.drop_column("task_queue", "status")
    op.drop_column("task_queue", "parser")
    op.drop_column("task_queue", "program")

    op.alter_column("task_queue", "created_on", nullable=False)
    op.alter_column("task_queue", "priority", nullable=False)

    # Service Queue
    op.drop_index("ix_service_queue_modified_on", table_name="service_queue")
    op.drop_index("ix_service_queue_priority", table_name="service_queue")
    op.drop_index("ix_service_queue_status", table_name="service_queue")
    op.drop_index("ix_service_queue_status_tag_hash", table_name="service_queue")
    op.create_index("ix_service_queue_tag", "service_queue", ["tag"], unique=False)
    op.execute(sa.text("CREATE INDEX ix_service_queue_waiting_sort ON service_queue (priority desc, created_on)"))
    op.drop_column("service_queue", "status")
    op.execute("DROP TYPE taskstatusenum")
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")

    # ### commands auto generated by Alembic - please adjust! ###
    # op.add_column('task_queue', sa.Column('program', sa.VARCHAR(), autoincrement=False, nullable=True))
    # op.add_column('task_queue', sa.Column('parser', sa.VARCHAR(), autoincrement=False, nullable=True))
    # op.add_column('task_queue', sa.Column('status', postgresql.ENUM('running', 'waiting', 'error', 'complete', name='taskstatusenum'), autoincrement=False, nullable=True))
    # op.add_column('task_queue', sa.Column('manager', sa.VARCHAR(), autoincrement=False, nullable=True))
    # op.add_column('task_queue', sa.Column('procedure', sa.VARCHAR(), autoincrement=False, nullable=True))
    # op.add_column('task_queue', sa.Column('modified_on', postgresql.TIMESTAMP(), autoincrement=False, nullable=True))
    # op.create_foreign_key('task_queue_manager_fkey', 'task_queue', 'queue_manager', ['manager'], ['name'], ondelete='SET NULL')
    # op.drop_index('ix_task_queue_tag', table_name='task_queue')
    # op.drop_index('ix_task_queue_required_programs', table_name='task_queue')
    # op.create_index('ix_task_waiting_sort', 'task_queue', ['priority', 'created_on'], unique=False)
    # op.create_index('ix_task_queue_manager', 'task_queue', ['manager'], unique=False)
    # op.create_index('ix_task_queue_keys', 'task_queue', ['status', 'program', 'procedure', 'tag'], unique=False)
    # op.create_index('ix_task_queue_created_on', 'task_queue', ['created_on'], unique=False)
    # op.drop_column('task_queue', 'required_programs')
    # op.add_column('service_queue', sa.Column('status', postgresql.ENUM('running', 'waiting', 'error', 'complete', name='taskstatusenum'), autoincrement=False, nullable=True))
    # op.drop_index('ix_service_queue_tag', table_name='service_queue')
    # op.create_index('ix_service_queue_status_tag_hash', 'service_queue', ['status', 'tag'], unique=False)
    # op.create_index('ix_service_queue_status', 'service_queue', ['status'], unique=False)
    # op.create_index('ix_service_queue_priority', 'service_queue', ['priority'], unique=False)
    # op.create_index('ix_service_queue_modified_on', 'service_queue', ['modified_on'], unique=False)
    # op.add_column('queue_manager', sa.Column('procedures', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True))
    # ### end Alembic commands ###



"""task spec is nullable

Revision ID: 7a60e93aed72
Revises: bf31e6366a10
Create Date: 2021-12-19 11:39:43.489411

"""

from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "7a60e93aed72"
down_revision = "bf31e6366a10"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("task_queue", "spec", existing_type=postgresql.BYTEA(), nullable=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("task_queue", "spec", existing_type=postgresql.BYTEA(), nullable=False)
    # ### end Alembic commands ###



"""
Pre-0.5 configuration file specification

This is needed for migrating the old config
"""

import os
from pathlib import Path
from typing import Optional

import yaml

try:
    from pydantic.v1 import Field, validator
except ImportError:
    from pydantic import Field, validator

from .config import ConfigBase, ConfigCommon


class OldDatabaseSettings(ConfigBase):
    """
    Postgres Database settings
    """

    port: int = Field(5432, description="The postgresql default port")
    host: str = Field(
        "localhost",
        description="Default location for the postgres server. If not localhost, qcfractal command lines cannot manage "
        "the instance.",
    )
    username: str = Field(None, description="The postgres username to default to.")
    password: str = Field(None, description="The postgres password for the give user.")
    directory: str = Field(
        None, description="The physical location of the QCFractal instance data, defaults to the root folder."
    )
    database_name: str = Field("qcfractal_default", description="The database name to connect to.")
    logfile: str = Field("qcfractal_postgres.log", description="The logfile to write postgres logs.")
    own: bool = Field(
        True,
        description="If own is True, QCFractal will control the database instance. If False "
        "Postgres will expect a booted server at the database specification.",
    )

    class Config(ConfigCommon):
        pass


class OldViewSettings(ConfigBase):
    """
    HDF5 view settings
    """

    enable: bool = Field(True, description="Enable frozen-views.")
    directory: str = Field(None, description="Location of frozen-view data. If None, defaults to base_folder/views.")


class OldFractalServerSettings(ConfigBase):
    """
    Fractal Server settings
    """

    name: str = Field("QCFractal Server", description="The QCFractal server default name.")
    port: int = Field(7777, description="The QCFractal default port.")

    # TODO: to be removed, handled by Ngnix
    compress_response: bool = Field(
        True, description="Compress REST responses or not, should be True unless behind a proxy."
    )
    allow_read: bool = Field(True, description="Always allows read access to record tables.")
    security: str = Field(
        None,
        description="Optional user authentication. Specify 'local' to enable "
        "authentication through locally stored usernames. "
        "User permissions may be manipulated through the ``qcfractal-server "
        "user`` CLI.",
    )

    query_limit: int = Field(1000, description="The maximum number of records to return per query.")
    logfile: Optional[str] = Field("qcfractal_server.log", description="The logfile to write server logs.")
    loglevel: str = Field("info", description="Level of logging to enable (debug, info, warning, error, critical)")
    cprofile: Optional[str] = Field(
        None, description="Enable profiling via cProfile, and output cprofile data to this path"
    )
    service_frequency: int = Field(60, description="The frequency to update the QCFractal services.")
    max_active_services: int = Field(20, description="The maximum number of concurrent active services.")
    heartbeat_frequency: int = Field(1800, description="The frequency (in seconds) to check the heartbeat of workers.")
    log_apis: bool = Field(
        False,
        description="True or False. Store API access in the Database. This is an advanced "
        "option for servers accessed by external users through QCPortal.",
    )
    geo_file_path: Optional[str] = Field(
        None,
        description="Geoip2 cites file path (.mmdb) for resolving IP addresses. Defaults to [base_folder]/GeoLite2-City.mmdb",
    )

    _default_geo_filename: str = "GeoLite2-City.mmdb"

    @validator("logfile")
    def check_basis(cls, v):
        if v == "None":
            v = None
        return v

    class Config(ConfigCommon):
        pass


class OldFractalConfig(ConfigBase):
    """
    Top level configuration headers and options for a QCFractal Configuration File
    """

    # class variable, not in the pydantic model
    _defaults_file_path: str = os.path.expanduser("~/.qca/qcfractal_defaults.yaml")

    base_folder: str = Field(
        os.path.expanduser("~/.qca/qcfractal"),
        description="The QCFractal base instance to attach to. " "Default will be your home directory",
    )
    database: OldDatabaseSettings = OldDatabaseSettings()
    view: OldViewSettings = OldViewSettings()
    fractal: OldFractalServerSettings = OldFractalServerSettings()

    class Config(ConfigCommon):
        pass

    def __init__(self, **kwargs):
        # If no base_folder provided, read it from ~/.qca/qcfractal_defaults.yaml (if it exists)
        # else, use the default base_folder
        if "base_folder" in kwargs:
            kwargs["base_folder"] = os.path.expanduser(kwargs["base_folder"])
        else:
            if Path(OldFractalConfig._defaults_file_path).exists():
                with open(OldFractalConfig._defaults_file_path, "r") as handle:
                    kwargs["base_folder"] = yaml.load(handle.read(), Loader=yaml.FullLoader)["default_base_folder"]

        super().__init__(**kwargs)

    @property
    def base_path(self):
        return Path(self.base_folder)

    @property
    def config_file_path(self):
        return self.base_path / "qcfractal_config.yaml"

    @property
    def database_path(self):
        if self.database.directory is None:
            return self.base_path / "postgres"
        else:
            return Path(os.path.expanduser(self.database.directory))

    def database_uri(self, safe: bool = True, database: str = None) -> str:
        uri = "postgresql://"
        if self.database.username is not None:
            uri += f"{self.database.username}:"

            if self.database.password is not None:
                if safe:
                    pw = "*******"
                else:
                    pw = self.database.password
                uri += pw

            uri += "@"

        uri += f"{self.database.host}:{self.database.port}/"

        if database is None:
            uri += self.database.database_name
        else:
            uri += database

        return uri

    @property
    def view_path(self):
        if self.view.directory is None:
            default_view_path = self.base_path / "views"
            default_view_path.mkdir(parents=False, exist_ok=True)
            return default_view_path
        else:
            return Path(os.path.expanduser(self.view.directory))

    def geo_file_path(self):
        if self.fractal.geo_file_path:
            return self.fractal.geo_file_path
        else:
            return os.path.join(self.base_folder, self.fractal._default_geo_filename)



"""clean up and migrate output store

Revision ID: 95e84c316d4d
Revises: 6f3de1040c37
Create Date: 2023-03-02 19:50:25.674718

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "95e84c316d4d"
down_revision = "6f3de1040c37"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("output_store", "data", new_column_name="old_data")
    op.alter_column("output_store", "compression", new_column_name="compression_type")
    op.add_column("output_store", sa.Column("data", sa.LargeBinary))
    op.execute(sa.text("ALTER TABLE output_store ALTER COLUMN data SET STORAGE EXTERNAL"))


def downgrade():
    raise RuntimeError("Cannot downgrade")
    pass



"""finalize migration of properties

Revision ID: 5a98e27e4c10
Revises: 95e84c316d4d
Create Date: 2023-03-03 14:57:04.765823

"""

import numpy as np
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from qcfractal.db_socket.column_types import MsgpackExt

# revision identifiers, used by Alembic.
revision = "5a98e27e4c10"
down_revision = "95e84c316d4d"
branch_labels = None
depends_on = None


def convert_numpy(obj):
    if isinstance(obj, dict):
        return {k: convert_numpy(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple, set)):
        return [convert_numpy(v) for v in obj]
    elif isinstance(obj, np.ndarray):
        if obj.shape:
            return obj.ravel().tolist()
        else:
            return obj.tolist()
    else:
        return obj


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    base_table = table(
        "base_record",
        column("id", sa.Integer),
        column("record_type", sa.String),
        column("extras", MsgpackExt),
        column("new_extras", postgresql.JSONB),
        column("new_properties", postgresql.JSONB),
    )

    singlepoint_table = table(
        "singlepoint_record",
        column("id", sa.Integer),
        column("properties", postgresql.JSONB),
        column("return_result", MsgpackExt),
    )

    conn = op.get_bind()
    session = Session(conn)

    # Find records that haven't been migrated
    # Migrations should always set new_extras to something
    base_records = session.query(base_table).where(base_table.c.new_extras.is_(None)).yield_per(200)

    for br in base_records:
        # Find singlepoint rec if it exists
        new_properties = {}

        # Move return_result and properties from singlepoint into base record propertiesAnd return result
        if br.record_type == "singlepoint":
            sp_rec = session.query(singlepoint_table).where(singlepoint_table.c.id == br.id).one()
            if sp_rec.return_result is not None:
                new_properties["return_result"] = convert_numpy(sp_rec.return_result)
            if sp_rec.properties is not None:
                new_properties.update(sp_rec.properties)

        # convert any numpy objects in extras
        extras = convert_numpy(br.extras)

        # Add qcvars from extras (and remove from extras)
        if extras is not None:
            qcvars = extras.pop("qcvars", {})
            if qcvars:
                new_properties.update({k.lower(): v for k, v in qcvars.items()})

        session.execute(
            sa.update(base_table)
            .values(extras=None, new_extras=extras, new_properties=new_properties)
            .where(base_table.c.id == br.id)
        )

    session.commit()

    op.drop_column("base_record", "extras")
    op.drop_column("singlepoint_record", "return_result")
    op.drop_column("singlepoint_record", "properties")
    op.alter_column("base_record", "new_extras", new_column_name="extras")
    op.alter_column("base_record", "new_properties", new_column_name="properties")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("CANNOT DOWNGRADE")
    # ### end Alembic commands ###



from qcportal.molecules import Molecule, MoleculeUploadOptions

from typing import Iterable, List, Tuple


def molecules_from_files(
    file_list: Iterable[Tuple[str, str]], options: MoleculeUploadOptions
) -> List[Tuple[str, Molecule]]:
    return [(fname, Molecule.from_file(fpath)) for fname, fpath in file_list]



import json

import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "bf4b379a6ce4"


def get_empty_keywords_id(conn):
    res = conn.execute(sa.text("SELECT id FROM keywords WHERE hash_index = 'bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f'"))
    return res.scalar()


def add_qc_spec(conn, program, driver, method, basis, keywords_id, protocols):
    if basis is None:
        basis = ""
    if keywords_id is None:
        keywords_id = get_empty_keywords_id(conn)

    # Remove protocol defaults
    if protocols.get("wavefunction", None) in [None, "none"]:
        protocols.pop("wavefunction", None)
    if protocols.get("stdout", True) is True:
        protocols.pop("stdout", None)

    e_corr = protocols.pop("error_correction", None)

    if e_corr is not None:
        if e_corr.get("default_policy", True) is True:
            e_corr.pop("default_policy", None)
        if e_corr.get("policies", {}) in [None, {}]:
            e_corr.pop("policies", None)

    if e_corr not in [None, {}]:
        protocols["error_correction"] = e_corr

    protocols_str = json.dumps(protocols)

    conn.execute(
        sa.text(
            """
               INSERT INTO qc_specification (program, driver, method, basis, keywords_id, protocols)
               VALUES (:program, :driver, :method, :basis, :keywords_id, (:protocols)::jsonb)
               ON CONFLICT DO NOTHING
               """
        ),
        parameters=dict(
            program=program, driver=driver, method=method, basis=basis, keywords_id=keywords_id, protocols=protocols_str
        ),
    )

    res = conn.execute(
        sa.text(
            """SELECT id FROM qc_specification
                   WHERE program = :program
                   AND driver = :driver
                   AND method = :method
                   AND basis = :basis
                   AND keywords_id = :keywords_id
                   AND protocols = (:protocols)::jsonb
                """
        ),
        parameters=dict(
            program=program, driver=driver, method=method, basis=basis, keywords_id=keywords_id, protocols=protocols_str
        ),
    )
    return res.scalar()


def add_opt_spec(conn, qc_specification_id, program, keywords, protocols):
    keywords_str = json.dumps(keywords)

    # Remove protocol defaults
    if protocols.get("trajectory", "all") in [None, "all"]:
        protocols.pop("trajectory", None)

    protocols_str = json.dumps(protocols)

    conn.execute(
        sa.text(
            """
           INSERT INTO optimization_specification (program, keywords, protocols, qc_specification_id)
           VALUES (:program, (:keywords)::jsonb, (:protocols)::jsonb, :qc_specification_id)
           ON CONFLICT DO NOTHING
           """
        ),
        parameters=dict(
            program=program, keywords=keywords_str, protocols=protocols_str, qc_specification_id=qc_specification_id
        ),
    )

    res = conn.execute(
        sa.text(
            """
                SELECT id FROM optimization_specification
                WHERE program = :program
                AND keywords = (:keywords)::jsonb
                AND protocols = (:protocols)::jsonb
                AND qc_specification_id = :qc_specification_id
                """
        ),
        parameters=dict(
            program=program, keywords=keywords_str, protocols=protocols_str, qc_specification_id=qc_specification_id
        ),
    )

    return res.scalar()



"""rename manager tables

Revision ID: 01eb7ec0d803
Revises: cabe9df168a5
Create Date: 2021-11-02 09:26:51.278977

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

# revision identifiers, used by Alembic.
revision = "01eb7ec0d803"
down_revision = "cabe9df168a5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.rename_table("queue_manager", "compute_manager")
    op.rename_table("queue_manager_logs", "compute_manager_logs")

    op.execute(sa.text("ALTER INDEX queue_manager_logs_pkey RENAME TO compute_manager_logs_pkey"))
    op.execute(sa.text("ALTER SEQUENCE queue_manager_logs_id_seq RENAME TO compute_manager_logs_id_seq"))

    op.execute(sa.text("ALTER INDEX queue_manager_pkey RENAME TO compute_manager_pkey"))
    op.execute(sa.text("ALTER SEQUENCE queue_manager_id_seq RENAME TO compute_manager_id_seq"))

    op.execute(sa.text("ALTER INDEX ix_queue_manager_modified_on RENAME TO ix_compute_manager_modified_on"))
    op.execute(sa.text("ALTER INDEX ix_queue_manager_status RENAME TO ix_compute_manager_status"))
    op.execute(sa.text("ALTER INDEX ix_queue_manager_log_timestamp RENAME TO ix_compute_manager_logs_timestamp"))

    op.execute(
        sa.text("ALTER TABLE compute_manager RENAME CONSTRAINT queue_manager_name_key TO ux_compute_manager_name")
    )
    op.execute(
        sa.text(
            "ALTER TABLE compute_manager_logs RENAME CONSTRAINT queue_manager_logs_manager_id_fkey TO compute_manager_logs_manager_id_fkey"
        )
    )

    # delete/rename some columns
    op.drop_column("compute_manager", "uuid")
    op.drop_column("compute_manager", "returned")

    # Add the rejected column
    op.add_column("compute_manager", sa.Column("rejected", sa.Integer(), nullable=True))

    # Make lots of columns not nullable
    op.execute(sa.text("UPDATE compute_manager SET completed = 0 WHERE completed IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET submitted = 0 WHERE submitted IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET failures = 0 WHERE failures IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET rejected = 0 WHERE rejected IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET total_worker_walltime = 0.0 WHERE total_worker_walltime IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET total_task_walltime = 0.0 WHERE total_task_walltime IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET active_tasks = 0 WHERE active_tasks IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET active_memory = 0.0 WHERE active_memory IS NULL"))
    op.execute(sa.text("UPDATE compute_manager SET active_cores = 0 WHERE active_cores IS NULL"))

    op.alter_column("compute_manager", "name", existing_type=sa.VARCHAR(), nullable=False)
    op.alter_column("compute_manager", "cluster", existing_type=sa.VARCHAR(), nullable=False)
    op.alter_column("compute_manager", "hostname", existing_type=sa.VARCHAR(), nullable=False)
    op.alter_column(
        "compute_manager", "completed", existing_type=sa.INTEGER(), nullable=False, new_column_name="successes"
    )
    op.alter_column(
        "compute_manager", "submitted", existing_type=sa.INTEGER(), nullable=False, new_column_name="claimed"
    )
    op.alter_column("compute_manager", "failures", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column(
        "compute_manager",
        "total_worker_walltime",
        existing_type=postgresql.DOUBLE_PRECISION(precision=53),
        nullable=False,
    )
    op.alter_column(
        "compute_manager",
        "total_task_walltime",
        existing_type=postgresql.DOUBLE_PRECISION(precision=53),
        nullable=False,
    )
    op.alter_column("compute_manager", "active_tasks", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column("compute_manager", "active_cores", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column(
        "compute_manager", "active_memory", existing_type=postgresql.DOUBLE_PRECISION(precision=53), nullable=False
    )
    op.alter_column(
        "compute_manager",
        "status",
        existing_type=postgresql.ENUM("active", "inactive", name="managerstatusenum"),
        nullable=False,
    )
    op.alter_column("compute_manager", "created_on", existing_type=postgresql.TIMESTAMP(), nullable=False)
    op.alter_column("compute_manager", "modified_on", existing_type=postgresql.TIMESTAMP(), nullable=False)

    op.add_column("compute_manager_logs", sa.Column("rejected", sa.Integer(), nullable=True))
    op.execute(sa.text("UPDATE compute_manager_logs SET completed = 0 WHERE completed IS NULL"))
    op.execute(sa.text("UPDATE compute_manager_logs SET submitted = 0 WHERE submitted IS NULL"))
    op.execute(sa.text("UPDATE compute_manager_logs SET failures = 0 WHERE failures IS NULL"))
    op.execute(sa.text("UPDATE compute_manager_logs SET rejected = 0 WHERE rejected IS NULL"))
    op.execute(
        sa.text("UPDATE compute_manager_logs SET total_worker_walltime = 0.0 WHERE total_worker_walltime IS NULL")
    )
    op.execute(sa.text("UPDATE compute_manager_logs SET total_task_walltime = 0.0 WHERE total_task_walltime IS NULL"))
    op.execute(sa.text("UPDATE compute_manager_logs SET active_tasks = 0 WHERE active_tasks IS NULL"))
    op.execute(sa.text("UPDATE compute_manager_logs SET active_memory = 0.0 WHERE active_memory IS NULL"))
    op.execute(sa.text("UPDATE compute_manager_logs SET active_cores = 0 WHERE active_cores IS NULL"))
    op.alter_column(
        "compute_manager_logs", "completed", existing_type=sa.INTEGER(), nullable=False, new_column_name="successes"
    )
    op.alter_column(
        "compute_manager_logs", "submitted", existing_type=sa.INTEGER(), nullable=False, new_column_name="claimed"
    )
    op.alter_column("compute_manager_logs", "failures", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column(
        "compute_manager_logs",
        "total_worker_walltime",
        existing_type=postgresql.DOUBLE_PRECISION(precision=53),
        nullable=False,
    )
    op.alter_column(
        "compute_manager_logs",
        "total_task_walltime",
        existing_type=postgresql.DOUBLE_PRECISION(precision=53),
        nullable=False,
    )
    op.alter_column("compute_manager_logs", "active_tasks", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column("compute_manager_logs", "active_cores", existing_type=sa.INTEGER(), nullable=False)
    op.alter_column(
        "compute_manager_logs", "active_memory", existing_type=postgresql.DOUBLE_PRECISION(precision=53), nullable=False
    )

    # Migrate tags
    op.add_column("compute_manager", sa.Column("tags", postgresql.ARRAY(sa.String())))

    # tag was originally a string
    # So trim any curly braces and split by comma, forming an array
    op.execute(
        sa.text(
            r"UPDATE compute_manager SET tags = regexp_split_to_array(trim(BOTH '{}' FROM LOWER(tag)), '\s*,\s*') WHERE tag IS NOT NULL"
        )
    )

    # now set previously null tags to wildcard
    op.execute(sa.text(r"UPDATE compute_manager SET tags = ARRAY['*'] WHERE tag IS NULL"))

    # Now make column nullable
    op.alter_column(
        "compute_manager",
        "tags",
        nullable=False,
    )

    op.alter_column(
        "compute_manager",
        "rejected",
        nullable=False,
    )

    op.alter_column(
        "compute_manager_logs",
        "rejected",
        nullable=False,
    )

    # Deal with null manager and qcengine version
    op.execute(sa.text(r"UPDATE compute_manager SET manager_version = 'v0' WHERE manager_version IS NULL"))

    op.execute(sa.text(r"UPDATE compute_manager SET qcengine_version = 'v0' WHERE qcengine_version IS NULL"))

    op.execute(sa.text(r"UPDATE compute_manager SET programs = '{}'::json WHERE programs IS NULL"))

    op.alter_column("compute_manager", "qcengine_version", nullable=False)
    op.alter_column("compute_manager", "manager_version", nullable=False)
    op.alter_column("compute_manager", "programs", nullable=False)

    op.drop_column("compute_manager", "tag")
    op.drop_column("compute_manager", "configuration")

    # Convert programs from list to dictionary
    programs_helper = table("compute_manager", column("id", sa.Integer), column("programs", sa.JSON))

    bind = op.get_bind()
    session = Session(bind=bind)

    for p in session.query(programs_helper).yield_per(500):
        if isinstance(p.programs, list):
            new_progs = {k.lower(): None for k in p.programs}
            session.execute(
                programs_helper.update().where(programs_helper.c.id == p.id).values({"programs": new_progs})
            )

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



from typing import List

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.exceptions import LimitExceededError
from qcportal.torsiondrive import (
    TorsiondriveDatasetSpecification,
    TorsiondriveDatasetNewEntry,
    TorsiondriveAddBody,
    TorsiondriveQueryFilters,
)
from qcportal.utils import calculate_limit


#####################
# Record
#####################


@api_v1.route("/records/torsiondrive/bulkCreate", methods=["POST"])
@wrap_global_route("records", "add")
def add_torsiondrive_records_v1(body_data: TorsiondriveAddBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.add_records
    if len(body_data.initial_molecules) > limit:
        raise LimitExceededError(
            f"Cannot add {len(body_data.initial_molecules)} torsiondrive records - limit is {limit}"
        )

    return storage_socket.records.torsiondrive.add(
        initial_molecules=body_data.initial_molecules,
        td_spec=body_data.specification,
        as_service=body_data.as_service,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/records/torsiondrive/<int:record_id>/optimizations", methods=["GET"])
@wrap_global_route("records", "read")
def get_torsiondrive_optimizations_v1(record_id: int):
    return storage_socket.records.torsiondrive.get_optimizations(record_id)


@api_v1.route("/records/torsiondrive/<int:record_id>/minimum_optimizations", methods=["GET"])
@wrap_global_route("records", "read")
def get_torsiondrive_minimum_optimizations_v1(record_id: int):
    return storage_socket.records.torsiondrive.get_minimum_optimizations(record_id)


@api_v1.route("/records/torsiondrive/<int:record_id>/initial_molecules", methods=["GET"])
@wrap_global_route("records", "read")
def get_torsiondrive_initial_molecules_ids_v1(record_id: int):
    return storage_socket.records.torsiondrive.get_initial_molecules_ids(record_id)


@api_v1.route("/records/torsiondrive/query", methods=["POST"])
@wrap_global_route("datasets", "read")
def query_torsiondrive_v1(body_data: TorsiondriveQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records
    body_data.limit = calculate_limit(max_limit, body_data.limit)

    return storage_socket.records.torsiondrive.query(body_data)


#####################
# Dataset
#####################


@api_v1.route("/datasets/torsiondrive/<int:dataset_id>/specifications", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_torsiondrive_dataset_specifications_v1(dataset_id: int, body_data: List[TorsiondriveDatasetSpecification]):
    return storage_socket.datasets.torsiondrive.add_specifications(dataset_id, body_data)


@api_v1.route("/datasets/torsiondrive/<int:dataset_id>/entries/bulkCreate", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_torsiondrive_dataset_entries_v1(dataset_id: int, body_data: List[TorsiondriveDatasetNewEntry]):
    return storage_socket.datasets.torsiondrive.add_entries(dataset_id, new_entries=body_data)


@api_v1.route("/datasets/torsiondrive/<int:dataset_id>/background_add_entries", methods=["POST"])
@wrap_global_route("datasets", "modify")
def background_add_torsiondrive_dataset_entries_v1(dataset_id: int, body_data: List[TorsiondriveDatasetNewEntry]):
    return storage_socket.datasets.torsiondrive.background_add_entries(dataset_id, new_entries=body_data)



"""add module to access log and remove access_type

Revision ID: b03545e6292a
Revises: 1638db43303c
Create Date: 2023-06-14 09:16:54.987767

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b03545e6292a"
down_revision = "1638db43303c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("access_log", sa.Column("module", sa.String(), nullable=True))
    op.create_index("ix_access_log_module", "access_log", ["module"], unique=False)

    op.execute(sa.text("UPDATE access_log SET module = 'compute' WHERE access_type LIKE '%api/v1/tasks%'"))
    op.execute(sa.text("UPDATE access_log SET module = 'auth' WHERE access_type LIKE '%auth/v1/tasks%'"))
    op.execute(sa.text("UPDATE access_log SET module = 'dashboard' WHERE access_type LIKE '%dashboard/v1%'"))
    op.execute(sa.text("UPDATE access_log SET module = 'old' WHERE access_type IS NULL"))
    op.drop_column("access_log", "access_type")
    op.alter_column("access_log", "access_date", new_column_name="timestamp")
    op.alter_column("access_log", "access_method", new_column_name="method")
    op.execute(sa.text("ALTER INDEX ix_access_log_access_date RENAME TO ix_access_log_timestamp"))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("access_log", sa.Column("access_type", sa.VARCHAR(), autoincrement=False, nullable=False))
    op.drop_index("ix_access_log_module", table_name="access_log")
    op.drop_column("access_log", "module")
    op.alter_column("access_log", "timestamp", new_column_name="access_date")
    op.alter_column("access_log", "method", new_column_name="access_method")
    op.execute(sa.text("ALTER INDEX ix_access_log_timestamp RENAME TO ix_access_log_access_date"))
    # ### end Alembic commands ###



"""rename base dataset extra

Revision ID: 21e29b68cbd5
Revises: b9b7b6926b8b
Create Date: 2022-05-22 10:21:59.093555

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "21e29b68cbd5"
down_revision = "b9b7b6926b8b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("base_dataset", "extra", new_column_name="extras")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("base_dataset", "extras", new_column_name="extra")
    # ### end Alembic commands ###



"""rename and refactor optimization

Revision ID: cdd9f3dc10e3
Revises: 160352419195
Create Date: 2021-12-08 14:53:27.286030

"""

import os
import sys

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers.v0_50_helpers import get_empty_keywords_id

# revision identifiers, used by Alembic.
revision = "cdd9f3dc10e3"
down_revision = "160352419195"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # The optimization specification table
    op.create_table(
        "optimization_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(length=100), nullable=False),
        sa.Column("qc_specification_id", sa.Integer(), nullable=False),
        sa.Column("keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("protocols", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.CheckConstraint("program = LOWER(program)", name="ck_optimization_specification_program_lower"),
        sa.ForeignKeyConstraint(
            ["qc_specification_id"],
            ["qc_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "program",
            "qc_specification_id",
            "keywords",
            "protocols",
            name="ux_optimization_specification_keys",
        ),
    )
    op.create_index("ix_optimization_specification_keywords", "optimization_specification", ["keywords"], unique=False)
    op.create_index("ix_optimization_specification_program", "optimization_specification", ["program"], unique=False)
    op.create_index(
        "ix_optimization_specification_protocols", "optimization_specification", ["protocols"], unique=False
    )
    op.create_index(
        "ix_optimization_specification_qc_specification_id",
        "optimization_specification",
        ["qc_specification_id"],
        unique=False,
    )

    # Optimization <-> Singlepoint association (for Trajectory)
    op.drop_constraint("opt_result_association_result_id_fkey", "opt_result_association", type_="foreignkey")
    op.drop_constraint("opt_result_association_opt_id_fkey", "opt_result_association", type_="foreignkey")
    op.alter_column("opt_result_association", "result_id", new_column_name="singlepoint_id")
    op.alter_column("opt_result_association", "opt_id", new_column_name="optimization_id")
    op.create_foreign_key(
        "optimization_trajectory_optimization_id_fkey",
        "opt_result_association",
        "optimization_procedure",
        ["optimization_id"],
        ["id"],
        ondelete="cascade",
    )
    op.create_foreign_key(
        "optimization_trajectory_singlepoint_id_fkey",
        "opt_result_association",
        "singlepoint_record",
        ["singlepoint_id"],
        ["id"],
    )

    # Now the optimization table
    op.execute(
        sa.text(
            "ALTER TABLE optimization_procedure RENAME CONSTRAINT optimization_procedure_id_fkey TO optimization_record_id_fkey"
        )
    )
    op.execute(sa.text("ALTER INDEX optimization_procedure_pkey RENAME TO optimization_record_pkey"))

    op.add_column("optimization_procedure", sa.Column("specification_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "optimization_record_specification_id_fkey",
        "optimization_procedure",
        "optimization_specification",
        ["specification_id"],
        ["id"],
    )

    op.drop_constraint("optimization_procedure_final_molecule_fkey", "optimization_procedure", type_="foreignkey")
    op.drop_constraint("optimization_procedure_initial_molecule_fkey", "optimization_procedure", type_="foreignkey")
    op.alter_column("optimization_procedure", "initial_molecule", new_column_name="initial_molecule_id")
    op.alter_column("optimization_procedure", "final_molecule", new_column_name="final_molecule_id")
    op.create_foreign_key(
        "optimization_record_initial_molecule_id_fkey",
        "optimization_procedure",
        "molecule",
        ["initial_molecule_id"],
        ["id"],
    )
    op.create_foreign_key(
        "optimization_record_final_molecule_id_fkey",
        "optimization_procedure",
        "molecule",
        ["final_molecule_id"],
        ["id"],
    )

    op.create_index(
        "ix_optimization_record_initial_molecule_id", "optimization_procedure", ["initial_molecule_id"], unique=False
    )
    op.create_index(
        "ix_optimization_record_final_molecule_id", "optimization_procedure", ["final_molecule_id"], unique=False
    )
    op.create_index(
        "ix_optimization_record_specification_id", "optimization_procedure", ["specification_id"], unique=False
    )

    # Migrate data to the specification tables
    # First, qcspec -> singlepoint specification

    # Now we can go with qc_spec
    empty_kw_id = get_empty_keywords_id(op.get_bind())

    # remove empty strings an json null
    op.execute(
        sa.text(
            r"UPDATE optimization_procedure SET qc_spec = (qc_spec::jsonb - 'keywords')::json WHERE qc_spec::jsonb->>'keywords' = ''"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE optimization_procedure SET qc_spec = (qc_spec::jsonb - 'keywords')::json WHERE qc_spec::jsonb->>'keywords' = 'null'"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE optimization_procedure SET qc_spec = (qc_spec::jsonb - 'keywords')::json WHERE qc_spec::jsonb->>'keywords' IS NULL"
        )
    )

    # Remove program from keywords
    op.execute(sa.text(r"UPDATE optimization_procedure SET keywords = (keywords::jsonb - 'program')::json"))

    # Protocols for qc_spec were always ignored. So set them with the default
    op.execute(
        sa.text(
            f"""
               INSERT INTO qc_specification (program, driver, method, basis, keywords_id, protocols)
               SELECT DISTINCT o.qc_spec->>'program',
                               'deferred'::singlepointdriver,
                               o.qc_spec->>'method',
                               COALESCE(o.qc_spec->>'basis', ''),
                               COALESCE((o.qc_spec->>'keywords')::int, {empty_kw_id}),
                               '{{}}'::jsonb
               FROM optimization_procedure o
               ON CONFLICT DO NOTHING
               """
        )
    )

    # Populate the optimization specification table
    op.execute(
        sa.text(
            r"""UPDATE base_record SET protocols = '{}'::json WHERE protocols::jsonb = '{"trajectory": "all"}'::jsonb"""
        )
    )

    op.execute(
        sa.text(
            f"""
               INSERT INTO optimization_specification (program, keywords, protocols, qc_specification_id)
               SELECT DISTINCT o.program,
                               o.keywords::jsonb,
                               br.protocols::jsonb,
                               (
                               SELECT id from qc_specification sp
                               WHERE sp.program = o.qc_spec->>'program'
                               AND sp.driver = 'deferred'::singlepointdriver
                               AND sp.method = o.qc_spec->>'method'
                               AND sp.basis = COALESCE(o.qc_spec->>'basis', '')
                               AND sp.keywords_id = COALESCE((o.qc_spec->>'keywords')::int, {empty_kw_id}) 
                               AND sp.protocols = '{{}}'::jsonb
                               )
               FROM optimization_procedure o INNER JOIN base_record br on br.id = o.id
               """
        )
    )

    # Now this mess
    op.execute(
        sa.text(
            f"""
               UPDATE optimization_procedure o
               SET specification_id = ss.id
               FROM optimization_specification ss, base_record as br
               WHERE o.id = br.id
               AND o.program = ss.program
               AND o.keywords::jsonb = ss.keywords
               AND br.protocols::jsonb = ss.protocols
               AND ss.qc_specification_id = (
                               SELECT id from qc_specification sp
                               WHERE sp.program = o.qc_spec->>'program'
                               AND sp.driver = 'deferred'::singlepointdriver
                               AND sp.method = o.qc_spec->>'method'
                               AND sp.basis = COALESCE(o.qc_spec->>'basis', '')
                               AND sp.keywords_id = COALESCE((o.qc_spec->>'keywords')::int, {empty_kw_id}) 
                               AND sp.protocols = '{{}}'::jsonb
                               )
               """
        )
    )

    op.drop_index("ix_optimization_program", table_name="optimization_procedure")
    op.drop_column("optimization_procedure", "qc_spec")
    op.drop_column("optimization_procedure", "keywords")
    op.drop_column("optimization_procedure", "schema_version")
    op.drop_column("optimization_procedure", "program")

    # Columns not nullable
    op.alter_column("optimization_procedure", "specification_id", nullable=False)

    # Rename the tables
    op.rename_table("opt_result_association", "optimization_trajectory")

    # Drop and recreate the pkey, since column names changed
    op.drop_constraint("opt_result_association_pkey", "optimization_trajectory", type_="primary")
    op.execute(
        sa.text("ALTER TABLE optimization_trajectory ADD PRIMARY KEY (optimization_id, singlepoint_id, position)")
    )

    op.rename_table("optimization_procedure", "optimization_record")

    # We can now finally remove the protocols from base_record
    op.drop_column("base_record", "protocols")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")



"""Migrate server info/stats/access tables

Revision ID: ae97b389022a
Revises: 05f7b00bfac8
Create Date: 2021-10-23 09:54:45.993070

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "ae97b389022a"
down_revision = "05f7b00bfac8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("access_log", "postal_code")
    op.drop_column("access_log", "country")

    op.alter_column(
        "access_log",
        "ip_address",
        existing_type=sa.VARCHAR(),
        type_=postgresql.INET(),
        existing_nullable=True,
        postgresql_using="ip_address::inet",
    )
    op.alter_column(
        "access_log", "country_code", existing_type=sa.VARCHAR(), type_=sa.CHAR(length=2), existing_nullable=True
    )
    op.alter_column(
        "access_log",
        "ip_lat",
        existing_type=sa.VARCHAR(),
        type_=sa.Float(),
        existing_nullable=True,
        postgresql_using="ip_lat::double precision",
    )
    op.alter_column(
        "access_log",
        "ip_long",
        existing_type=sa.VARCHAR(),
        type_=sa.Float(),
        existing_nullable=True,
        postgresql_using="ip_long::double precision",
    )

    op.alter_column(
        "server_stats_log",
        "kvstore_count",
        new_column_name="outputstore_count",
        existing_type=sa.INTEGER(),
        type_=sa.BigInteger(),
        existing_nullable=True,
    )

    op.alter_column(
        "server_stats_log", "molecule_count", existing_type=sa.INTEGER(), type_=sa.BigInteger(), existing_nullable=True
    )
    op.alter_column(
        "server_stats_log", "result_count", existing_type=sa.INTEGER(), type_=sa.BigInteger(), existing_nullable=True
    )
    op.alter_column(
        "server_stats_log", "access_count", existing_type=sa.INTEGER(), type_=sa.BigInteger(), existing_nullable=True
    )
    op.alter_column(
        "server_stats_log", "error_count", existing_type=sa.INTEGER(), type_=sa.BigInteger(), existing_nullable=True
    )

    # Rename kv_store in the table information column
    # (for simplicity, just do a text search and replace)
    op.execute(
        sa.text(
            "UPDATE server_stats_log SET db_table_information = replace(db_table_information::text, 'kv_store', 'output_store')::json"
        )
    )

    op.alter_column("access_log", "access_date", existing_type=postgresql.TIMESTAMP(), nullable=False)
    op.alter_column("internal_error_log", "error_date", existing_type=postgresql.TIMESTAMP(), nullable=False)
    op.alter_column("internal_error_log", "qcfractal_version", existing_type=sa.VARCHAR(), nullable=False)

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""Add trigger for internal jobs notification

Revision ID: 1abf80db6c19
Revises: 64cae90ab9fe
Create Date: 2023-06-29 17:42:51.293964

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1abf80db6c19"
down_revision = "64cae90ab9fe"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(
        sa.text(
            """
    CREATE OR REPLACE FUNCTION public.qca_internal_jobs_notify()
    RETURNS trigger
    LANGUAGE plpgsql
    AS $function$
        BEGIN
          PERFORM pg_notify('check_internal_jobs', '');
          RETURN NEW;
        END
        $function$
    ;
        """
        )
    )

    op.execute(
        sa.text(
            "CREATE TRIGGER qca_internal_jobs_insert_tr AFTER INSERT ON public.internal_jobs FOR EACH ROW EXECUTE FUNCTION qca_internal_jobs_notify();"
        )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("DROP TRIGGER qca_internal_jobs_insert_tr ON public.internal_jobs;"))
    op.execute(sa.text("DROP FUNCTION public.qca_internal_jobs_notify();"))
    # ### end Alembic commands ###



"""Remove keywords index and add hashing

Revision ID: 981f69781d65
Revises: 148fef89c2ec
Create Date: 2022-10-12 15:34:38.809659

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "981f69781d65"
down_revision = "148fef89c2ec"
branch_labels = None
depends_on = None


from migration_helpers.hashing import hash_dict_1


def create_hashes(table):
    conn = op.get_bind()
    res = conn.execute(sa.text(f"SELECT id,keywords FROM {table};"))
    all_kw = res.fetchall()

    for spec_id, kw in all_kw:
        h = hash_dict_1(kw)
        op.execute(sa.text(f"""UPDATE {table} SET keywords_hash = '{h}' WHERE id = {spec_id};"""))


def upgrade():
    op.execute(sa.text("DROP INDEX IF EXISTS ix_qc_specification_keywords"))
    op.execute(sa.text("ALTER TABLE qc_specification DROP CONSTRAINT IF EXISTS ux_qc_specification_keys"))
    op.add_column("qc_specification", sa.Column("keywords_hash", sa.String()))
    op.drop_index("ix_qc_specification_protocols")
    op.create_unique_constraint(
        "ux_qc_specification_keys",
        "qc_specification",
        ["program", "driver", "method", "basis", "keywords_hash", "protocols"],
    )
    create_hashes("qc_specification")
    op.alter_column("qc_specification", "keywords_hash", nullable=False)

    op.execute(sa.text("DROP INDEX IF EXISTS ix_optimization_specification_keywords"))
    op.execute(
        sa.text("ALTER TABLE optimization_specification DROP CONSTRAINT IF EXISTS ux_optimization_specification_keys")
    )
    op.drop_index("ix_optimization_specification_protocols")
    op.add_column("optimization_specification", sa.Column("keywords_hash", sa.String()))
    op.create_unique_constraint(
        "ux_optimization_specification_keys",
        "optimization_specification",
        ["program", "qc_specification_id", "keywords_hash", "protocols"],
    )
    create_hashes("optimization_specification")
    op.alter_column("optimization_specification", "keywords_hash", nullable=False)

    op.execute(sa.text("DROP INDEX IF EXISTS ix_torsiondrive_specification_keywords"))
    op.execute(
        sa.text("ALTER TABLE torsiondrive_specification DROP CONSTRAINT IF EXISTS ux_torsiondrive_specification_keys")
    )
    op.add_column("torsiondrive_specification", sa.Column("keywords_hash", sa.String()))
    op.create_unique_constraint(
        "ux_torsiondrive_specification_keys",
        "torsiondrive_specification",
        ["program", "optimization_specification_id", "keywords_hash"],
    )
    create_hashes("torsiondrive_specification")
    op.alter_column("torsiondrive_specification", "keywords_hash", nullable=False)

    op.execute(sa.text("DROP INDEX IF EXISTS ix_gridoptimization_specification_keywords"))
    op.execute(
        sa.text(
            "ALTER TABLE gridoptimization_specification DROP CONSTRAINT IF EXISTS ux_gridoptimization_specification_keys"
        )
    )
    op.add_column("gridoptimization_specification", sa.Column("keywords_hash", sa.String()))
    op.create_unique_constraint(
        "ux_gridoptimization_specification_keys",
        "gridoptimization_specification",
        ["program", "optimization_specification_id", "keywords_hash"],
    )
    create_hashes("gridoptimization_specification")
    op.alter_column("gridoptimization_specification", "keywords_hash", nullable=False)

    op.execute(sa.text("DROP INDEX IF EXISTS ix_manybody_specification_keywords"))
    op.execute(sa.text("ALTER TABLE manybody_specification DROP CONSTRAINT IF EXISTS ux_manybody_specification_keys"))
    op.add_column("manybody_specification", sa.Column("keywords_hash", sa.String()))
    op.create_unique_constraint(
        "ux_manybody_specification_keys",
        "manybody_specification",
        ["program", "singlepoint_specification_id", "keywords_hash"],
    )
    create_hashes("manybody_specification")
    op.alter_column("manybody_specification", "keywords_hash", nullable=False)

    op.execute(sa.text("DROP INDEX IF EXISTS ix_reaction_specification_keywords"))
    op.execute(sa.text("ALTER TABLE reaction_specification DROP CONSTRAINT IF EXISTS ux_reaction_specification_keys"))
    op.add_column("reaction_specification", sa.Column("keywords_hash", sa.String()))
    op.create_unique_constraint(
        "ux_reaction_specification_keys",
        "reaction_specification",
        ["program", "singlepoint_specification_id", "optimization_specification_id", "keywords_hash"],
    )
    create_hashes("reaction_specification")
    op.alter_column("reaction_specification", "keywords_hash", nullable=False)

    op.execute(sa.text("DROP INDEX IF EXISTS ix_neb_specification_keywords"))
    op.execute(sa.text("ALTER TABLE neb_specification DROP CONSTRAINT IF EXISTS ux_neb_specification_keys"))
    op.add_column("neb_specification", sa.Column("keywords_hash", sa.String()))
    op.create_unique_constraint(
        "ux_neb_specification_keys", "neb_specification", ["program", "singlepoint_specification_id", "keywords_hash"]
    )
    create_hashes("neb_specification")
    op.alter_column("neb_specification", "keywords_hash", nullable=False)


def downgrade():
    raise RuntimeError("CANNOT DOWNGRADE")



"""Add defaults to keywords

Revision ID: 6128f0430433
Revises: 77baa72171b9
Create Date: 2022-09-29 10:23:54.435874

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6128f0430433"
down_revision = "77baa72171b9"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        sa.text(
            """
        UPDATE torsiondrive_specification
        SET keywords = (keywords || jsonb_build_object(
             'dihedrals', COALESCE(keywords->'dihedrals', '[]'::jsonb),
             'grid_spacing', COALESCE(keywords->'grid_spacing', '[]'::jsonb),
             'dihedral_ranges', COALESCE(keywords->'dihedral_ranges', Null::jsonb),
             'energy_decrease_thresh', COALESCE(keywords->'energy_decrease_thresh', Null::jsonb),
             'energy_upper_limit', COALESCE(keywords->'energy_upper_limit', Null::jsonb))
             )
        """
        )
    )

    op.execute(
        sa.text(
            """
        UPDATE gridoptimization_specification
        SET keywords = (keywords || jsonb_build_object(
             'scans', COALESCE(keywords->'scans', '[]'::jsonb),
             'preoptimization', COALESCE(keywords->'preoptimization', 'true'::jsonb))
             )
        """
        )
    )

    op.execute(
        sa.text(
            """
        UPDATE manybody_specification
        SET keywords = (keywords || jsonb_build_object('max_nbody', COALESCE(keywords->'max_nbody', Null::jsonb)))
        """
        )
    )


def downgrade():
    pass



"""compress wavefunctions

Revision ID: 6f3de1040c37
Revises: 1a61b3bb1ee4
Create Date: 2023-02-09 10:00:49.665770

"""

import sqlalchemy as sa
from alembic import op
from qcelemental.models.results import WavefunctionProperties
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from qcfractal.db_socket.column_types import MsgpackExt
from qcportal.compression import compress, CompressionEnum

# revision identifiers, used by Alembic.
revision = "6f3de1040c37"
down_revision = "1a61b3bb1ee4"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.add_column("wavefunction_store", sa.Column("compression_type", sa.Enum(CompressionEnum, nullable=True)))
    op.add_column("wavefunction_store", sa.Column("compression_level", sa.Integer, nullable=True))
    op.add_column("wavefunction_store", sa.Column("data", sa.LargeBinary, nullable=True))

    op.execute(sa.text("ALTER TABLE native_file ALTER COLUMN data SET STORAGE EXTERNAL"))

    # Migrate wavefunction data
    old_wfn_table = table(
        "wavefunction_store",
        column("id", sa.Integer),
        column("record_id", sa.Integer),
        column("basis", MsgpackExt),
        column("restricted", sa.Boolean),
        column("h_core_a", MsgpackExt),
        column("h_core_b", MsgpackExt),
        column("h_effective_a", MsgpackExt),
        column("h_effective_b", MsgpackExt),
        column("scf_orbitals_a", MsgpackExt),
        column("scf_orbitals_b", MsgpackExt),
        column("scf_density_a", MsgpackExt),
        column("scf_density_b", MsgpackExt),
        column("scf_fock_a", MsgpackExt),
        column("scf_fock_b", MsgpackExt),
        column("scf_eigenvalues_a", MsgpackExt),
        column("scf_eigenvalues_b", MsgpackExt),
        column("scf_occupations_a", MsgpackExt),
        column("scf_occupations_b", MsgpackExt),
        column("orbitals_a", sa.String),
        column("orbitals_b", sa.String),
        column("density_a", sa.String),
        column("density_b", sa.String),
        column("fock_a", sa.String),
        column("fock_b", sa.String),
        column("eigenvalues_a", sa.String),
        column("eigenvalues_b", sa.String),
        column("occupations_a", sa.String),
        column("occupations_b", sa.String),
    )

    conn = op.get_bind()
    session = Session(conn)

    old_wfns = session.query(old_wfn_table).yield_per(200)
    for old_wfn in old_wfns:
        wfn_dict = dict(
            basis=old_wfn.basis,
            restricted=old_wfn.restricted,
            h_core_a=old_wfn.h_core_a,
            h_core_b=old_wfn.h_core_b,
            h_effective_a=old_wfn.h_effective_a,
            h_effective_b=old_wfn.h_effective_b,
            scf_orbitals_a=old_wfn.scf_orbitals_a,
            scf_orbitals_b=old_wfn.scf_orbitals_b,
            scf_density_a=old_wfn.scf_density_a,
            scf_density_b=old_wfn.scf_density_b,
            scf_fock_a=old_wfn.scf_fock_a,
            scf_fock_b=old_wfn.scf_fock_b,
            scf_eigenvalues_a=old_wfn.scf_eigenvalues_a,
            scf_eigenvalues_b=old_wfn.scf_eigenvalues_b,
            scf_occupations_a=old_wfn.scf_occupations_a,
            scf_occupations_b=old_wfn.scf_occupations_b,
            orbitals_a=old_wfn.orbitals_a,
            orbitals_b=old_wfn.orbitals_b,
            density_a=old_wfn.density_a,
            density_b=old_wfn.density_b,
            fock_a=old_wfn.fock_a,
            fock_b=old_wfn.fock_b,
            eigenvalues_a=old_wfn.eigenvalues_a,
            eigenvalues_b=old_wfn.eigenvalues_b,
            occupations_a=old_wfn.occupations_a,
            occupations_b=old_wfn.occupations_b,
        )

        # prune None values
        wfn_dict = {k: v for k, v in wfn_dict.items() if v is not None}
        wfn_prop = WavefunctionProperties(**wfn_dict)

        wfn_compressed, ctype, clevel = compress(wfn_prop.dict(encoding="json"), CompressionEnum.zstd)

        r = conn.execute(
            sa.text(
                """UPDATE wavefunction_store SET
                     h_core_a=NULL,
                     h_core_b=NULL,
                     h_effective_a=NULL,
                     h_effective_b=NULL,
                     scf_orbitals_a=NULL,
                     scf_orbitals_b=NULL,
                     scf_density_a=NULL,
                     scf_density_b=NULL,
                     scf_fock_a=NULL,
                     scf_fock_b=NULL,
                     scf_eigenvalues_a=NULL,
                     scf_eigenvalues_b=NULL,
                     scf_occupations_a=NULL,
                     scf_occupations_b=NULL,
                     orbitals_a=NULL,
                     orbitals_b=NULL,
                     density_a=NULL,
                     density_b=NULL,
                     fock_a=NULL,
                     fock_b=NULL,
                     eigenvalues_a=NULL,
                     eigenvalues_b=NULL,
                     occupations_a=NULL,
                     occupations_b=NULL,
                     compression_type='zstd',
                     compression_level=:clevel,
                     data=:cdata
                   WHERE id = :id"""
            ),
            parameters={"cdata": wfn_compressed, "clevel": clevel, "id": old_wfn.id},
        )

    op.drop_column("wavefunction_store", "basis")
    op.drop_column("wavefunction_store", "restricted")
    op.drop_column("wavefunction_store", "h_core_a")
    op.drop_column("wavefunction_store", "h_core_b")
    op.drop_column("wavefunction_store", "h_effective_a")
    op.drop_column("wavefunction_store", "h_effective_b")
    op.drop_column("wavefunction_store", "scf_orbitals_a")
    op.drop_column("wavefunction_store", "scf_orbitals_b")
    op.drop_column("wavefunction_store", "scf_density_a")
    op.drop_column("wavefunction_store", "scf_density_b")
    op.drop_column("wavefunction_store", "scf_fock_a")
    op.drop_column("wavefunction_store", "scf_fock_b")
    op.drop_column("wavefunction_store", "scf_eigenvalues_a")
    op.drop_column("wavefunction_store", "scf_eigenvalues_b")
    op.drop_column("wavefunction_store", "scf_occupations_a")
    op.drop_column("wavefunction_store", "scf_occupations_b")
    op.drop_column("wavefunction_store", "orbitals_a")
    op.drop_column("wavefunction_store", "orbitals_b")
    op.drop_column("wavefunction_store", "density_a")
    op.drop_column("wavefunction_store", "density_b")
    op.drop_column("wavefunction_store", "fock_a")
    op.drop_column("wavefunction_store", "fock_b")
    op.drop_column("wavefunction_store", "eigenvalues_a")
    op.drop_column("wavefunction_store", "eigenvalues_b")
    op.drop_column("wavefunction_store", "occupations_a")
    op.drop_column("wavefunction_store", "occupations_b")

    op.alter_column("wavefunction_store", "compression_type", nullable=False)
    op.alter_column("wavefunction_store", "compression_level", nullable=False)
    op.alter_column("wavefunction_store", "data", nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")
    # ### end Alembic commands ###



"""Add check constraints for tags

Revision ID: 37bf1ec2b1d1
Revises: 45b5ec1ed88b
Create Date: 2022-04-03 10:50:30.660492

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import column, func


# revision identifiers, used by Alembic.
revision = "37bf1ec2b1d1"
down_revision = "45b5ec1ed88b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("UPDATE task_queue SET tag = LOWER(tag)"))
    op.execute(sa.text("UPDATE service_queue SET tag = LOWER(tag)"))
    op.execute(sa.text("UPDATE compute_manager SET tags = LOWER(tags::text)::text[]"))
    op.execute(sa.text("UPDATE compute_manager SET programs = LOWER(programs::text)::json"))

    op.create_check_constraint(
        "ck_task_queue_tag_lower",
        "task_queue",
        column("tag").cast(sa.TEXT) == func.lower(column("tag").cast(sa.TEXT)),
    )

    op.create_check_constraint(
        "ck_service_queue_tag_lower",
        "service_queue",
        column("tag").cast(sa.TEXT) == func.lower(column("tag").cast(sa.TEXT)),
    )

    op.create_check_constraint(
        "ck_compute_manager_programs_lower",
        "compute_manager",
        column("programs").cast(sa.TEXT) == func.lower(column("programs").cast(sa.TEXT)),
    )

    op.create_check_constraint(
        "ck_compute_manager_tags_lower",
        "compute_manager",
        column("tags").cast(sa.TEXT) == func.lower(column("tags").cast(sa.TEXT)),
    )
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""migrate molecule table

Revision ID: fd76a1459fa3
Revises: 2a334994fcb3
Create Date: 2021-10-12 10:31:53.993770

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "fd76a1459fa3"
down_revision = "2a334994fcb3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "molecule", "molecule_hash", existing_type=sa.VARCHAR(), type_=sa.CHAR(length=40), existing_nullable=True
    )
    op.alter_column(
        "molecule",
        "identifiers",
        existing_type=postgresql.JSON(astext_type=sa.Text()),
        type_=postgresql.JSONB(astext_type=sa.Text()),
        existing_nullable=True,
    )

    op.alter_column("molecule", "fix_com", existing_type=sa.BOOLEAN(), nullable=False)
    op.alter_column("molecule", "fix_orientation", existing_type=sa.BOOLEAN(), nullable=False)

    op.create_index("ix_molecule_identifiers", "molecule", ["identifiers"], unique=False, postgresql_using="gin")
    op.drop_column("molecule", "molecular_formula")
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""Add full uri to access log

Revision ID: 5d1989f8b59f
Revises: bba52fa8efbe
Create Date: 2022-03-28 15:17:23.125656

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5d1989f8b59f"
down_revision = "bba52fa8efbe"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("access_log", sa.Column("full_uri", sa.String(), nullable=True))
    op.drop_column("access_log", "extra_params")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("access_log", sa.Column("extra_params", sa.VARCHAR(), autoincrement=False, nullable=True))
    op.drop_column("access_log", "full_uri")
    # ### end Alembic commands ###



"""move geolocation

Revision ID: f9d784e816d2
Revises: 698131e6ffd5
Create Date: 2023-05-16 08:41:51.101634

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f9d784e816d2"
down_revision = "698131e6ffd5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "server_stats_metadata",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("date_value", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )

    # Reset internal jobs - will be recreated on server startup
    op.execute(sa.text("DELETE FROM internal_jobs WHERE status IN ('running', 'waiting')"))

    op.execute(
        sa.text(
            """INSERT INTO server_stats_metadata (name, date_value)
                          VALUES ('last_geolocated_date', now() AT TIME ZONE 'utc')
                     """
        )
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("server_stats_metadata")
    # ### end Alembic commands ###



"""rename and refactor singlepoint

Revision ID: 160352419195
Revises: 88e6b7d536d0
Create Date: 2021-11-12 13:43:15.488344

"""

import os
import sys

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers.v0_50_helpers import get_empty_keywords_id

# revision identifiers, used by Alembic.
revision = "160352419195"
down_revision = "88e6b7d536d0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    driver_enum = postgresql.ENUM(
        "energy", "gradient", "hessian", "properties", "deferred", name="singlepointdriver", create_type=True
    )

    ########################################
    # Create qc_specification table
    op.create_table(
        "qc_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(length=100), nullable=False),
        sa.Column("driver", driver_enum, nullable=False),
        sa.Column("method", sa.String(length=100), nullable=False),
        sa.Column("basis", sa.String(length=100), nullable=False),
        sa.Column("keywords_id", sa.Integer(), nullable=False),
        sa.Column("protocols", postgresql.JSONB(), nullable=False),
        sa.CheckConstraint("basis = LOWER(basis)", name="ck_qc_specification_basis_lower"),
        sa.CheckConstraint("method = LOWER(method)", name="ck_qc_specification_method_lower"),
        sa.CheckConstraint("program = LOWER(program)", name="ck_qc_specification_program_lower"),
        sa.ForeignKeyConstraint(
            ["keywords_id"],
            ["keywords.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "program", "driver", "method", "basis", "keywords_id", "protocols", name="ux_qc_specification_keys"
        ),
    )
    op.create_index("ix_qc_specification_program", "qc_specification", ["program"], unique=False)
    op.create_index("ix_qc_specification_driver", "qc_specification", ["driver"], unique=False)
    op.create_index("ix_qc_specification_method", "qc_specification", ["method"], unique=False)
    op.create_index("ix_qc_specification_basis", "qc_specification", ["basis"], unique=False)
    op.create_index("ix_qc_specification_keywords_id", "qc_specification", ["keywords_id"], unique=False)
    op.create_index("ix_qc_specification_protocols", "qc_specification", ["protocols"], unique=False)

    ########################################
    # Add columns to the specification table, and rename some columns
    op.add_column("result", sa.Column("specification_id", sa.Integer(), nullable=True))  # Temporarily nullable
    op.create_index("ix_singlepoint_record_specification_id", "result", ["specification_id"], unique=False)
    op.create_foreign_key(
        "singlepoint_record_specification_id_fkey", "result", "qc_specification", ["specification_id"], ["id"]
    )

    # op.drop_constraint("result_molecule_fkey", "result", type_="foreignkey")
    op.alter_column("result", "molecule", new_column_name="molecule_id")
    op.alter_column(
        "result",
        "properties",
        existing_type=postgresql.JSON(astext_type=sa.Text()),
        type_=postgresql.JSONB(astext_type=sa.Text()),
        existing_nullable=True,
    )

    op.execute(
        sa.text("ALTER TABLE result RENAME CONSTRAINT result_molecule_fkey TO singlepoint_record_molecule_id_fkey")
    )

    ########################################
    # Now for the data migration
    # Adding the empty keyword surrogate
    op.execute(
        sa.text(
            """INSERT INTO keywords (hash_index, values, lowercase, exact_floats, comments)
               VALUES ('bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f',
                   '{}'::json,
                   true,
                   false,
                   'Empty keywords - use program defaults'
                   )
               ON CONFLICT DO NOTHING;"""
        )
    )

    empty_kw_id = get_empty_keywords_id(op.get_bind())

    ########################################
    # Update protocols to remove defaults and null
    # Column is not nullable, but sometimes stores json null
    op.execute(
        sa.text(
            r"""
               UPDATE base_record
               SET protocols = '{}'::jsonb
               WHERE protocols = 'null'::jsonb
               """
        )
    )

    # Remove default wavefunction
    op.execute(
        sa.text(
            r"""
               UPDATE base_record
               SET protocols = (protocols - 'wavefunction')
               WHERE record_type = 'singlepoint'
               AND protocols @> '{"wavefunction": "none"}'::jsonb;
               """
        )
    )

    # Remove default stdout
    op.execute(
        sa.text(
            r"""
               UPDATE base_record
               SET protocols = (protocols - 'stdout')
               WHERE record_type = 'singlepoint'
               AND protocols @> '{"stdout": true}'::jsonb;
               """
        )
    )

    # Remove default error correction protocol
    op.execute(
        sa.text(
            r"""
               UPDATE base_record
               SET protocols = (protocols #- '{error_correction,default_policy}')
               WHERE record_type = 'singlepoint'
               AND protocols @> '{"error_correction": {"default_policy": true}}'::jsonb
               """
        )
    )

    op.execute(
        sa.text(
            r"""
               UPDATE base_record
               SET protocols = (protocols #- '{error_correction,policies}')
               WHERE record_type = 'singlepoint'
               AND protocols->'error_correction'->'policies' = 'null'::jsonb
               OR protocols->'error_correction'->'policies' = '{}'::jsonb;
               """
        )
    )

    op.execute(
        sa.text(
            r"""
               UPDATE base_record
               SET protocols = (protocols - 'error_correction')
               WHERE record_type = 'singlepoint'
               AND protocols->'error_correction' = '{}'::jsonb;
               """
        )
    )

    ########################################
    # Populate the qc_specification table
    # Coalesce null basis set, keywords, protocols into something not null
    op.execute(
        sa.text(
            f"""
               INSERT INTO qc_specification (program, driver, method, basis, keywords_id, protocols)
               SELECT DISTINCT r.program,
                               r.driver::singlepointdriver,
                               r.method,
                               COALESCE(r.basis, ''),
                               COALESCE(r.keywords, {empty_kw_id}),
                               COALESCE(br.protocols, '{{}}'::jsonb)
               FROM result r INNER JOIN base_record br on r.id = br.id;
               """
        )
    )

    op.execute(
        sa.text(
            f"""
              UPDATE result AS r
              SET specification_id = ss.id
              FROM qc_specification AS ss, base_record as br
              WHERE r.id = br.id
              AND ss.program = r.program
              AND ss.driver = r.driver::singlepointdriver
              AND ss.method = r.method
              AND ss.basis = COALESCE(r.basis, '')
              AND ss.keywords_id = COALESCE(r.keywords, {empty_kw_id})
              AND ss.protocols = COALESCE(br.protocols, '{{}}'::jsonb);
            """
        )
    )

    ##############################################
    # Add & populate foreign keys on wavefunction store
    op.add_column("wavefunction_store", sa.Column("record_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "wavefunction_store_record_id_fkey", "wavefunction_store", "result", ["record_id"], ["id"], ondelete="CASCADE"
    )
    op.create_unique_constraint("ux_wavefunction_store_record_id", "wavefunction_store", ["record_id"])

    op.execute(
        sa.text(
            f"""
              UPDATE wavefunction_store SET record_id = result.id
              FROM result
              WHERE result.wavefunction_data_id = wavefunction_store.id
            """
        )
    )

    # Populate some columns of wavefunction based on the "wavefunction" dictionary column of result
    op.execute(
        sa.text(
            f"""
              UPDATE wavefunction_store SET
              orbitals_a = result.wavefunction->'return_map'->>'orbitals_a'::text,
              orbitals_b = result.wavefunction->'return_map'->>'orbitals_b',
              density_a = result.wavefunction->'return_map'->>'density_a',
              density_b = result.wavefunction->'return_map'->>'density_b',
              fock_a = result.wavefunction->'return_map'->>'fock_a',
              fock_b = result.wavefunction->'return_map'->>'fock_b',
              eigenvalues_a = result.wavefunction->'return_map'->>'eigenvalues_a',
              eigenvalues_b = result.wavefunction->'return_map'->>'eigenvalues_b',
              occupations_a = result.wavefunction->'return_map'->>'occupations_a',
              occupations_b = result.wavefunction->'return_map'->>'occupations_b'
              FROM result
              WHERE result.wavefunction_data_id = wavefunction_store.id
            """
        )
    )

    ########################################
    # rename some indices and constraints
    op.drop_index("ix_results_molecule", table_name="result")
    op.create_index("ix_singlepoint_record_molecule_id", "result", ["molecule_id"], unique=False)

    op.execute(sa.text("ALTER TABLE result RENAME CONSTRAINT result_id_fkey TO singlepoint_record_id_fkey"))

    # Now drop the unused columns/constraints/indices
    op.drop_constraint("result_wavefunction_data_id_fkey", "result", type_="foreignkey")
    op.drop_constraint("uix_results_keys", "result", type_="unique")
    op.drop_constraint("result_keywords_fkey", "result", type_="foreignkey")
    op.drop_column("result", "program")
    op.drop_column("result", "driver")
    op.execute(sa.text("DROP TYPE driverenum"))
    op.drop_column("result", "method")
    op.drop_column("result", "basis")
    op.drop_column("result", "keywords")
    op.drop_column("result", "wavefunction_data_id")
    op.drop_column("result", "wavefunction")

    # Drop any orphan wavefunction
    op.execute(sa.text("DELETE FROM wavefunction_store WHERE record_id IS NULL"))

    # Now make stuff not nullable
    op.alter_column("result", "specification_id", nullable=False)
    op.alter_column("wavefunction_store", "record_id", nullable=False)

    # Now rename the table
    op.rename_table("result", "singlepoint_record")
    op.execute(sa.text("ALTER INDEX result_pkey RENAME TO singlepoint_record_pkey"))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")



"""refactoring gridoptimization

Revision ID: 0f6d9e6ef312
Revises: bb4804242cea
Create Date: 2022-01-03 15:02:26.005753

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from qcfractal.db_socket.column_types import PlainMsgpackExt

# revision identifiers, used by Alembic.
revision = "0f6d9e6ef312"
down_revision = "bb4804242cea"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # Grid optimization spec table
    op.create_table(
        "gridoptimization_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(length=100), nullable=False),
        sa.Column("optimization_specification_id", sa.Integer(), nullable=False),
        sa.Column("keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.CheckConstraint("program = LOWER(program)", name="ck_gridoptimization_specification_program_lower"),
        sa.ForeignKeyConstraint(
            ["optimization_specification_id"],
            ["optimization_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "program", "optimization_specification_id", "keywords", name="ux_gridoptimization_specification_keys"
        ),
    )

    op.create_index(
        "ix_gridoptimization_specification_keywords", "gridoptimization_specification", ["keywords"], unique=False
    )
    op.create_index(
        "ix_gridoptimization_specification_optimization_specification_id",
        "gridoptimization_specification",
        ["optimization_specification_id"],
        unique=False,
    )
    op.create_index(
        "ix_gridoptimization_specification_program", "gridoptimization_specification", ["program"], unique=False
    )

    # Modify the optimization association table
    op.alter_column("grid_optimization_association", "opt_id", new_column_name="optimization_id", nullable=False)
    op.alter_column("grid_optimization_association", "grid_opt_id", new_column_name="gridoptimization_id")

    op.drop_constraint("grid_optimization_association_opt_id_fkey", "grid_optimization_association", type_="foreignkey")
    op.drop_constraint(
        "grid_optimization_association_grid_opt_id_fkey", "grid_optimization_association", type_="foreignkey"
    )
    op.create_foreign_key(
        "gridoptimization_optimizations_optimization_id_fkey",
        "grid_optimization_association",
        "optimization_record",
        ["optimization_id"],
        ["id"],
    )
    op.drop_constraint("grid_optimization_association_pkey", "grid_optimization_association", type_="primary")
    op.create_primary_key(
        "gridoptimization_optimizations_pkey",
        "grid_optimization_association",
        ["gridoptimization_id", "key"],
    )

    # Now the procedure table
    op.add_column("grid_optimization_procedure", sa.Column("specification_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "gridoptimization_record_specification_id_fkey",
        "grid_optimization_procedure",
        "gridoptimization_specification",
        ["specification_id"],
        ["id"],
    )

    op.drop_constraint("grid_optimization_procedure_id_fkey", "grid_optimization_procedure", type_="foreignkey")
    op.create_foreign_key(
        "gridoptimization_record_id_fkey",
        "grid_optimization_procedure",
        "base_record",
        ["id"],
        ["id"],
        ondelete="cascade",
    )

    op.drop_constraint(
        "grid_optimization_procedure_starting_molecule_fkey", "grid_optimization_procedure", type_="foreignkey"
    )
    op.drop_constraint(
        "grid_optimization_procedure_initial_molecule_fkey", "grid_optimization_procedure", type_="foreignkey"
    )
    op.alter_column("grid_optimization_procedure", "initial_molecule", new_column_name="initial_molecule_id")
    op.alter_column("grid_optimization_procedure", "starting_molecule", new_column_name="starting_molecule_id")
    op.create_foreign_key(
        "gridoptimization_record_initial_molecule_id_fkey",
        "grid_optimization_procedure",
        "molecule",
        ["initial_molecule_id"],
        ["id"],
    )
    op.create_foreign_key(
        "gridoptimization_record_starting_molecule_id_fkey",
        "grid_optimization_procedure",
        "molecule",
        ["starting_molecule_id"],
        ["id"],
    )

    op.drop_constraint("grid_optimization_procedure_pkey", "grid_optimization_procedure", type_="primary")
    op.create_primary_key(
        "gridoptimization_record_pkey",
        "grid_optimization_procedure",
        ["id"],
    )

    # Needed to wait for re-creating the primary key
    op.create_foreign_key(
        "gridoptimization_optimizations_gridoptimization_id_fkey",
        "grid_optimization_association",
        "grid_optimization_procedure",
        ["gridoptimization_id"],
        ["id"],
        ondelete="cascade",
    )

    ###########################################################
    # NOW THE BIG MIGRATION
    ###########################################################
    op.execute("ALTER TABLE grid_optimization_procedure ALTER COLUMN qc_spec TYPE JSONB")
    op.execute("ALTER TABLE grid_optimization_procedure ALTER COLUMN optimization_spec TYPE JSONB")
    op.execute("ALTER TABLE grid_optimization_procedure ALTER COLUMN keywords TYPE JSONB")

    # The empty, default keywords
    res = op.get_bind().execute(
        sa.text("SELECT id FROM keywords WHERE hash_index = 'bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f'")
    )
    empty_kw = res.scalar()

    # Fiddle with the specifications
    # First, a hack. The MolSSI database has some old data with a non-existent hash still there. Replace
    # that with the appropriate keyword
    op.execute(
        sa.text(
            r"""UPDATE grid_optimization_procedure SET qc_spec = (qc_spec || '{"keywords": "2"}') WHERE qc_spec->>'keywords' = '5c954fa6b6a2de5f188ea234'"""
        )
    )

    # Remove empty and null keywords
    op.execute(
        sa.text(
            r"UPDATE grid_optimization_procedure SET qc_spec = (qc_spec - 'keywords') WHERE qc_spec->>'keywords' = ''"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE grid_optimization_procedure SET qc_spec = (qc_spec - 'keywords') WHERE qc_spec->>'keywords' = 'null'"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE grid_optimization_procedure SET qc_spec = (qc_spec - 'keywords') WHERE qc_spec->>'keywords' IS NULL"
        )
    )

    # Insert the qcspec
    # Protocols for qc_spec were always ignored. So set them with the default
    op.execute(
        sa.text(
            f"""
               INSERT INTO qc_specification (program, driver, method, basis, keywords_id, protocols)
               SELECT DISTINCT go.qc_spec->>'program',
                               'deferred'::singlepointdriver,
                               go.qc_spec->>'method',
                               COALESCE(go.qc_spec->>'basis', ''),
                               COALESCE((go.qc_spec->>'keywords')::int, {empty_kw}),
                               '{{}}'::jsonb
               FROM grid_optimization_procedure go
               ON CONFLICT DO NOTHING
               """
        )
    )

    # remove explicitly specified default protocols
    op.execute(
        sa.text(
            r"""UPDATE grid_optimization_procedure SET optimization_spec = (optimization_spec || '{"protocols": "{}"}') WHERE optimization_spec->'protocols' = '{"trajectory": "all"}'"""
        )
    )

    # Now the optimization_spec
    # Remove empty and null keywords
    op.execute(
        sa.text(
            r"UPDATE grid_optimization_procedure SET optimization_spec = (optimization_spec - 'keywords') WHERE optimization_spec->>'keywords' = ''"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE grid_optimization_procedure SET optimization_spec = (optimization_spec - 'keywords') WHERE optimization_spec->>'keywords' = 'null'"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE grid_optimization_procedure SET optimization_spec = (optimization_spec - 'keywords') WHERE optimization_spec->>'keywords' IS NULL"
        )
    )

    op.execute(
        sa.text(
            f"""
               INSERT INTO optimization_specification (program, keywords, protocols, qc_specification_id)
               SELECT DISTINCT go.optimization_spec->>'program',
                               COALESCE(go.optimization_spec->'keywords', '{{}}'),
                               COALESCE(go.optimization_spec->'protocols', '{{}}'),
                               (
                               SELECT id from qc_specification sp
                               WHERE sp.program = go.qc_spec->>'program'
                               AND sp.driver = 'deferred'::singlepointdriver
                               AND sp.method = go.qc_spec->>'method'
                               AND sp.basis = COALESCE(go.qc_spec->>'basis', '')
                               AND sp.keywords_id = COALESCE((go.qc_spec->>'keywords')::int, {empty_kw})
                               AND sp.protocols = '{{}}'
                               )
               FROM grid_optimization_procedure go
               ON CONFLICT DO NOTHING
               """
        )
    )

    # And the gridoptimization spec
    op.execute(
        sa.text(
            f"""
               INSERT INTO gridoptimization_specification (program, keywords, optimization_specification_id)
               SELECT DISTINCT 'gridoptimization',
                               go.keywords,
                               (
                                    SELECT id from optimization_specification os
                                    WHERE os.program = go.optimization_spec->>'program'
                                    AND os.keywords = COALESCE(go.optimization_spec->'keywords', '{{}}')
                                    AND os.protocols = COALESCE(go.optimization_spec->'protocols', '{{}}')
                                    AND os.qc_specification_id =
                                       (
                                           SELECT id from qc_specification sp
                                           WHERE sp.program = go.qc_spec->>'program'
                                           AND sp.driver = 'deferred'::singlepointdriver
                                           AND sp.method = go.qc_spec->>'method'
                                           AND sp.basis = COALESCE(go.qc_spec->>'basis', '')
                                           AND sp.keywords_id = COALESCE((go.qc_spec->>'keywords')::int, {empty_kw})
                                           AND sp.protocols = '{{}}'
                                       )
                                )
               FROM grid_optimization_procedure go
               """
        )
    )

    # Now add this to the gridoptimization spec column
    op.execute(
        sa.text(
            f"""
               UPDATE grid_optimization_procedure go
               SET specification_id = (
                   SELECT id FROM gridoptimization_specification gs
                   WHERE gs.program = 'gridoptimization'
                   AND gs.keywords = go.keywords
                   AND gs.optimization_specification_id = (
                             SELECT id from optimization_specification os
                             WHERE os.program = go.optimization_spec->>'program'
                             AND os.keywords = COALESCE(go.optimization_spec->'keywords', '{{}}')
                             AND os.protocols = COALESCE(go.optimization_spec->'protocols', '{{}}')
                             AND os.qc_specification_id = (
                                    SELECT id from qc_specification sp
                                    WHERE sp.program = go.qc_spec->>'program'
                                    AND sp.driver = 'deferred'::singlepointdriver
                                    AND sp.method = go.qc_spec->>'method'
                                    AND sp.basis = COALESCE(go.qc_spec->>'basis', '')
                                    AND sp.keywords_id = COALESCE((go.qc_spec->>'keywords')::int, {empty_kw})
                                    AND sp.protocols = '{{}}'
                            )
                        )
                   )
               """
        )
    )

    # Now migrate the service queue
    # Temporary ORM
    service_table = table(
        "service_queue",
        column("id", sa.Integer),
        column("service_state", PlainMsgpackExt),
    )

    bind = op.get_bind()
    session = Session(bind=bind)

    services = session.query(service_table).all()

    for service in services:
        if "torsiondrive_state" in service.service_state:
            continue

        # We have a gridoptimization
        # Remove the optimization template
        service.service_state.pop("optimization_template")
        session.execute(
            service_table.update()
            .where(service_table.c.id == service.id)
            .values({"service_state": service.service_state})
        )

    # rename "initial_opt" to "preoptimization" in the service dependencies
    op.execute(
        sa.text(
            """UPDATE service_dependencies SET extras = extras || '{"key": "preoptimization"}'
                          WHERE extras->>'key' = 'initial_opt'"""
        )
    )

    # Make columns not nullable now that they are populated
    op.alter_column("grid_optimization_procedure", "specification_id", nullable=False)

    # Drop unused columns
    op.drop_column("grid_optimization_procedure", "optimization_spec")
    op.drop_column("grid_optimization_procedure", "keywords")
    op.drop_column("grid_optimization_procedure", "qc_spec")
    op.drop_column("grid_optimization_procedure", "final_energy_dict")

    # Finally rename the tables
    op.rename_table("grid_optimization_association", "gridoptimization_optimizations")
    op.rename_table("grid_optimization_procedure", "gridoptimization_record")

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""migrate keywords constraints

Revision ID: d36e2ec09f27
Revises: 651954cbdec6
Create Date: 2021-11-03 12:05:20.152697

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = "d36e2ec09f27"
down_revision = "651954cbdec6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_keywords_hash_index", table_name="keywords")
    op.create_unique_constraint("ux_keywords_hash_index", "keywords", ["hash_index"])

    op.alter_column("keywords", "values", existing_type=postgresql.JSON(astext_type=sa.Text()), nullable=False)
    op.alter_column("keywords", "lowercase", existing_type=sa.BOOLEAN(), nullable=False)
    op.alter_column("keywords", "exact_floats", existing_type=sa.BOOLEAN(), nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("keywords", "exact_floats", existing_type=sa.BOOLEAN(), nullable=True)
    op.alter_column("keywords", "lowercase", existing_type=sa.BOOLEAN(), nullable=True)
    op.alter_column("keywords", "values", existing_type=postgresql.JSON(astext_type=sa.Text()), nullable=True)
    op.drop_constraint("ux_keywords_hash_index", "keywords", type_="unique")
    op.create_index("ix_keywords_hash_index", "keywords", ["hash_index"], unique=False)
    # ### end Alembic commands ###



# Used in some migrations before changing how roles work

default_roles = {
    "admin": {
        "Statement": [
            {"Effect": "Allow", "Action": "*", "Resource": "*"},
        ]
    },
    "maintain": {
        "Statement": [
            {"Effect": "Allow", "Action": "*", "Resource": "*"},
        ]
    },
    "read": {
        "Statement": [
            {"Effect": "Allow", "Action": "READ", "Resource": "*"},
            {"Effect": "Allow", "Action": "WRITE", "Resource": ["/api/v1/users", "/api/v1/me"]},
            {
                "Effect": "Deny",
                "Action": "*",
                "Resource": [
                    "/api/v1/roles",
                    "/api/v1/managers",
                    "/api/v1/server_errors",
                    "/api/v1/access_logs",
                    "/api/v1/tasks",
                    "/api/v1/internal_jobs",
                ],
            },
        ]
    },
    "monitor": {
        "Statement": [
            {"Effect": "Allow", "Action": "READ", "Resource": "*"},
            {"Effect": "Allow", "Action": "WRITE", "Resource": "/api/v1/users"},
            {"Effect": "Deny", "Action": "*", "Resource": ["/api/v1/roles"]},
        ]
    },
    "compute": {
        "Statement": [
            {"Effect": "Allow", "Action": ["READ"], "Resource": "/api/v1/information"},
            {"Effect": "Allow", "Action": ["READ"], "Resource": "/compute/v1/information"},
            {"Effect": "Allow", "Action": ["READ", "WRITE"], "Resource": "/api/v1/users"},
            {"Effect": "Allow", "Action": "*", "Resource": ["/compute/v1/managers", "/compute/v1/tasks"]},
        ]
    },
    "submit": {
        "Statement": [
            {"Effect": "Allow", "Action": "READ", "Resource": "*"},
            {"Effect": "Allow", "Action": "WRITE", "Resource": "/api/v1/users"},
            {
                "Effect": "Deny",
                "Action": "*",
                "Resource": [
                    "/api/v1/roles",
                    "/api/v1/managers",
                    "/api/v1/server_errors",
                    "/api/v1/access_logs",
                    "/api/v1/tasks",
                    "/api/v1/internal_jobs",
                ],
            },
            {
                "Effect": "Allow",
                "Action": "*",
                "Resource": ["/api/v1/records", "/api/v1/molecules", "/api/v1/keywords", "/api/v1/datasets"],
            },
        ]
    },
}



"""Add service iteration record

Revision ID: ee6defa818d6
Revises: 11a500398106
Create Date: 2023-01-18 15:23:51.842564

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ee6defa818d6"
down_revision = "11a500398106"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "service_subtask_record",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("required_programs", sa.JSON(), nullable=False),
        sa.Column("function", sa.String(), nullable=False),
        sa.Column("function_kwargs", sa.JSON(), nullable=False),
        sa.Column("results", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["id"], ["base_record.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("service_subtask_record")
    # ### end Alembic commands ###



"""Clean up cascading deletes and add triggers

Revision ID: e645774b4c62
Revises: c23b3ba12869
Create Date: 2023-03-17 11:14:04.032459

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e645774b4c62"
down_revision = "c23b3ba12869"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("compute_manager_log_manager_id_fkey", "compute_manager_log", type_="foreignkey")
    op.create_foreign_key(None, "compute_manager_log", "compute_manager", ["manager_id"], ["id"], ondelete="cascade")

    op.execute(
        sa.text(
            """
    CREATE OR REPLACE FUNCTION public.qca_base_record_delete()
    RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
        BEGIN
          DELETE FROM base_record WHERE base_record.id = OLD.id;
          RETURN OLD;
        END
        $_$
    ;
        """
        )
    )

    op.execute(
        sa.text(
            "CREATE TRIGGER qca_gridoptimization_record_delete_base_tr AFTER DELETE ON public.gridoptimization_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )
    op.execute(
        sa.text(
            "CREATE TRIGGER qca_manybody_record_delete_base_tr AFTER DELETE ON public.manybody_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )
    op.execute(
        sa.text(
            "CREATE TRIGGER qca_neb_record_delete_base_tr AFTER DELETE ON public.neb_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )
    op.execute(
        sa.text(
            "CREATE TRIGGER qca_optimization_record_delete_base_tr AFTER DELETE ON public.optimization_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )
    op.execute(
        sa.text(
            "CREATE TRIGGER qca_reaction_record_delete_base_tr AFTER DELETE ON public.reaction_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )
    op.execute(
        sa.text(
            "CREATE TRIGGER qca_singlepoint_record_delete_base_tr AFTER DELETE ON public.singlepoint_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )
    op.execute(
        sa.text(
            "CREATE TRIGGER qca_torsiondrive_record_delete_base_tr AFTER DELETE ON public.torsiondrive_record FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"
        )
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "compute_manager_log", type_="foreignkey")
    op.create_foreign_key(
        "compute_manager_log_manager_id_fkey", "compute_manager_log", "compute_manager", ["manager_id"], ["id"]
    )
    # ### end Alembic commands ###



"""Add internal job queue

Revision ID: 64d8c7cc3a18
Revises: 9c91ede09098
Create Date: 2022-09-12 16:06:24.175386

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "64d8c7cc3a18"
down_revision = "9c91ede09098"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "internal_jobs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "status",
            sa.Enum("complete", "waiting", "running", "error", "cancelled", name="internaljobstatusenum"),
            nullable=False,
        ),
        sa.Column("added_date", sa.DateTime(), nullable=False),
        sa.Column("scheduled_date", sa.DateTime(), nullable=False),
        sa.Column("started_date", sa.DateTime(), nullable=True),
        sa.Column("last_updated", sa.DateTime(), nullable=True),
        sa.Column("ended_date", sa.DateTime(), nullable=True),
        sa.Column("runner_hostname", sa.String(), nullable=True),
        sa.Column("runner_uuid", sa.String(), nullable=True),
        sa.Column("progress", sa.Integer(), nullable=False),
        sa.Column("function", sa.String(), nullable=False),
        sa.Column("kwargs", sa.JSON(), nullable=False),
        sa.Column("after_function", sa.String(), nullable=True),
        sa.Column("after_function_kwargs", sa.JSON(), nullable=True),
        sa.Column("result", sa.JSON(), nullable=True),
        sa.Column("user", sa.String(), nullable=True),
        sa.Column("unique_name", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("unique_name", name="ux_internal_jobs_unique_name"),
    )
    op.create_index("ix_internal_jobs_added_date", "internal_jobs", ["added_date"], unique=False)
    op.create_index("ix_internal_jobs_last_updated", "internal_jobs", ["last_updated"], unique=False)
    op.create_index("ix_internal_jobs_name", "internal_jobs", ["name"], unique=False)
    op.create_index("ix_internal_jobs_scheduled_date", "internal_jobs", ["scheduled_date"], unique=False)
    op.create_index("ix_internal_jobs_status", "internal_jobs", ["status"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_internal_jobs_status", table_name="internal_jobs")
    op.drop_index("ix_internal_jobs_scheduled_date", table_name="internal_jobs")
    op.drop_index("ix_internal_jobs_name", table_name="internal_jobs")
    op.drop_index("ix_internal_jobs_last_updated", table_name="internal_jobs")
    op.drop_index("ix_internal_jobs_added_date", table_name="internal_jobs")
    op.drop_table("internal_jobs")
    # ### end Alembic commands ###



"""Add user auth types

Revision ID: 5db09491c69d
Revises: b1d1a21895ba
Create Date: 2022-07-07 16:30:42.581429

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5db09491c69d"
down_revision = "b1d1a21895ba"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    sa_enum = sa.Enum("password", name="authtypeenum")
    sa_enum.create(op.get_bind())

    op.add_column("user", sa.Column("auth_type", sa.Enum(name="authtypeenum"), nullable=True))
    op.execute(sa.text("UPDATE \"user\" SET auth_type = 'password'"))
    op.alter_column("user", "auth_type", nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("user", "auth_type")
    # ### end Alembic commands ###



"""Optimization spec hashing

Revision ID: e04bf1eedfbc
Revises: 6b758fd53ff0
Create Date: 2024-12-17 10:50:37.251828

"""

import sqlalchemy as sa
from alembic import op

from migration_helpers.hashing import hash_dict_1

# revision identifiers, used by Alembic.
revision = "e04bf1eedfbc"
down_revision = "6b758fd53ff0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("optimization_specification", sa.Column("specification_hash", sa.String(), nullable=True))

    # Generate hashes for existing specs
    conn = op.get_bind()
    res = conn.execute(sa.text(f"SELECT id, program, keywords, protocols FROM optimization_specification;"))
    all_spec = res.fetchall()
    for spec_id, program, keywords, protocols in all_spec:
        d = {
            "program": program,
            "keywords": keywords,
            "protocols": protocols,
        }
        h = hash_dict_1(d)
        op.execute(
            sa.text(f"""UPDATE optimization_specification SET specification_hash = '{h}' WHERE id = {spec_id};""")
        )

    op.alter_column("optimization_specification", "specification_hash", nullable=False)
    op.drop_constraint("ux_optimization_specification_keys", "optimization_specification", type_="unique")
    op.drop_column("optimization_specification", "keywords_hash")
    op.create_unique_constraint(
        "ux_optimization_specification_keys",
        "optimization_specification",
        ["specification_hash", "qc_specification_id"],
    )
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Downgrade not supported.")



"""change server stats collected

Revision ID: 00f7a74dcc37
Revises: e1bfb46d1055
Create Date: 2021-04-25 10:36:07.834628

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "00f7a74dcc37"
down_revision = "e1bfb46d1055"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("server_stats_log", sa.Column("error_count", sa.Integer(), nullable=True))
    op.add_column("server_stats_log", sa.Column("service_queue_status", sa.JSON(), nullable=True))
    op.add_column("server_stats_log", sa.Column("task_queue_status", sa.JSON(), nullable=True))
    op.drop_column("server_stats_log", "result_states")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "server_stats_log",
        sa.Column("result_states", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
    )
    op.drop_column("server_stats_log", "task_queue_status")
    op.drop_column("server_stats_log", "service_queue_status")
    op.drop_column("server_stats_log", "error_count")
    # ### end Alembic commands ###



"""cascade deletes to service queue

Revision ID: 1c01369d0bae
Revises: 0f6d9e6ef312
Create Date: 2022-01-28 10:44:03.394150

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "1c01369d0bae"
down_revision = "0f6d9e6ef312"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("service_queue_record_id_fkey", "service_queue", type_="foreignkey")
    op.create_foreign_key(None, "service_queue", "base_record", ["record_id"], ["id"], ondelete="cascade")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "service_queue", type_="foreignkey")
    op.create_foreign_key("service_queue_record_id_fkey", "service_queue", "base_record", ["record_id"], ["id"])
    # ### end Alembic commands ###



"""Fix more indices in compute manager log

Revision ID: f65aad4015cb
Revises: a27802e0d8fe
Create Date: 2023-06-25 10:33:38.594008

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "f65aad4015cb"
down_revision = "a27802e0d8fe"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_compute_manager_log_timestamp", table_name="compute_manager_log")
    op.create_index("ix_compute_manager_log_manager_id", "compute_manager_log", ["manager_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_compute_manager_log_manager_id", table_name="compute_manager_log")
    op.create_index("ix_compute_manager_log_timestamp", "compute_manager_log", ["timestamp"], unique=False)
    # ### end Alembic commands ###



"""add record info backup table

Revision ID: 7fb1093c30ca
Revises: cdd9f3dc10e3
Create Date: 2021-12-11 09:14:04.907844

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = "7fb1093c30ca"
down_revision = "cdd9f3dc10e3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    status_enum = postgresql.ENUM(
        "complete",
        "waiting",
        "running",
        "error",
        "cancelled",
        "deleted",
        "invalid",
        name="recordstatusenum",
        create_type=False,
    )

    op.create_table(
        "record_info_backup",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.Column("old_status", status_enum, nullable=False),
        sa.Column("old_tag", sa.String(), nullable=True),
        sa.Column("old_priority", sa.Integer(), nullable=True),
        sa.Column("modified_on", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["record_id"], ["base_record.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("record_info_backup")
    # ### end Alembic commands ###



"""Change manager stats

Revision ID: 2c48247776d0
Revises: 84e421577a1a
Create Date: 2023-03-27 18:44:13.468201

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "2c48247776d0"
down_revision = "84e421577a1a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("compute_manager", "total_worker_walltime")
    op.drop_column("compute_manager_log", "total_worker_walltime")

    op.alter_column("compute_manager", "total_task_walltime", new_column_name="total_cpu_hours")
    op.alter_column("compute_manager_log", "total_task_walltime", new_column_name="total_cpu_hours")
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""Remove versions table

Revision ID: 9a86ebfcfd1b
Revises: 5d1989f8b59f
Create Date: 2022-03-29 10:28:46.346932

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "9a86ebfcfd1b"
down_revision = "5d1989f8b59f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("versions")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "versions",
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column("created_on", postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
        sa.Column("elemental_version", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column("fractal_version", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.PrimaryKeyConstraint("id", name="versions_pkey"),
    )
    # ### end Alembic commands ###



"""Add internal_error_log table

Revision ID: e1bfb46d1055
Revises: 9e817a4a5cec
Create Date: 2021-04-24 13:57:02.485403

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e1bfb46d1055"
down_revision = "9e817a4a5cec"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "internal_error_log",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("error_date", sa.DateTime(), nullable=True),
        sa.Column("qcfractal_version", sa.String(), nullable=True),
        sa.Column("error_text", sa.String(), nullable=True),
        sa.Column("user", sa.String(), nullable=True),
        sa.Column("request_path", sa.String(), nullable=True),
        sa.Column("request_headers", sa.String(), nullable=True),
        sa.Column("request_body", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_internal_error_log_error_date"), "internal_error_log", ["error_date"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_internal_error_log_error_date"), table_name="internal_error_log")
    op.drop_table("internal_error_log")
    # ### end Alembic commands ###



"""Add qcengine to required programs

Revision ID: b1d1a21895ba
Revises: 97e1d19796c7
Create Date: 2022-07-07 09:17:33.447320

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b1d1a21895ba"
down_revision = "97e1d19796c7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("UPDATE task_queue SET required_programs = ARRAY['qcengine'] || required_programs"))
    op.execute(
        sa.text(
            """UPDATE compute_manager SET programs = programs::jsonb || jsonb_build_object('qcengine', qcengine_version)"""
        )
    )
    op.drop_column("compute_manager", "qcengine_version")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")
    # ### end Alembic commands ###



"""Add find_existing to services

Revision ID: e6f5053c7600
Revises: 1abf80db6c19
Create Date: 2023-08-13 17:01:08.736187

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e6f5053c7600"
down_revision = "1abf80db6c19"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("service_queue", sa.Column("find_existing", sa.Boolean(), nullable=True))
    op.execute(sa.text("UPDATE service_queue SET find_existing = true;"))
    op.alter_column("service_queue", "find_existing", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("service_queue", "find_existing")
    # ### end Alembic commands ###



"""make_reaction_service_flexible

Revision ID: c3cbccb2d572
Revises: f26c30450fdd
Create Date: 2022-05-12 14:56:15.632423

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "c3cbccb2d572"
down_revision = "f26c30450fdd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    ####################################################
    # Safe to assume we don't have anything to migrate
    # (since this is happening between major versions,
    #  and the previous version didn't have reactions)
    ####################################################

    op.create_table(
        "reaction_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(), nullable=False),
        sa.Column("singlepoint_specification_id", sa.Integer(), nullable=True),
        sa.Column("optimization_specification_id", sa.Integer(), nullable=True),
        sa.Column("keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.CheckConstraint("program = LOWER(program)", name="ck_reaction_specification_program_lower"),
        sa.CheckConstraint(
            "singlepoint_specification_id IS NOT NULL OR optimization_specification_id IS NOT NULL",
            name="ck_reaction_specification_specs",
        ),
        sa.ForeignKeyConstraint(
            ["optimization_specification_id"],
            ["optimization_specification.id"],
        ),
        sa.ForeignKeyConstraint(
            ["singlepoint_specification_id"],
            ["qc_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "singlepoint_specification_id",
            "optimization_specification_id",
            "keywords",
            name="ux_reaction_specification_keys",
        ),
    )
    op.create_index("ix_reaction_specification_keywords", "reaction_specification", ["keywords"], unique=False)
    op.create_index(
        "ix_reaction_specification_optimization_specification_id",
        "reaction_specification",
        ["optimization_specification_id"],
        unique=False,
    )
    op.create_index("ix_reaction_specification_program", "reaction_specification", ["program"], unique=False)
    op.create_index(
        "ix_reaction_specification_singlepoint_specification_id",
        "reaction_specification",
        ["singlepoint_specification_id"],
        unique=False,
    )
    op.add_column("reaction_component", sa.Column("coefficient", postgresql.DOUBLE_PRECISION(), nullable=False))
    op.add_column("reaction_component", sa.Column("optimization_id", sa.Integer(), nullable=True))
    op.alter_column("reaction_component", "singlepoint_id", existing_type=sa.INTEGER(), nullable=True)
    op.create_index("ix_reaction_component_optimization_id", "reaction_component", ["optimization_id"], unique=False)
    op.create_index("ix_reaction_component_singlepoint_id", "reaction_component", ["singlepoint_id"], unique=False)
    op.drop_constraint("reaction_component_reaction_id_molecule_id_fkey", "reaction_component", type_="foreignkey")
    op.create_foreign_key(None, "reaction_component", "optimization_record", ["optimization_id"], ["id"])
    op.drop_constraint("reaction_record_specification_id_fkey", "reaction_record", type_="foreignkey")
    op.create_foreign_key(None, "reaction_record", "reaction_specification", ["specification_id"], ["id"])
    op.drop_table("reaction_stoichiometry")
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""clean up dataset db model

Revision ID: 5a5105707147
Revises: 0b31ab4244de
Create Date: 2022-05-18 09:19:34.511867

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5a5105707147"
down_revision = "0b31ab4244de"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("collection", "collection_type", new_column_name="dataset_type")
    op.drop_index("ix_collection_lname", table_name="collection")
    op.drop_index("ix_collection_type", table_name="collection")
    op.create_index("ix_dataset_type", "collection", ["dataset_type"], unique=False)
    op.create_unique_constraint("uix_dataset_type_lname", "collection", ["dataset_type", "lname"])
    op.drop_column("collection", "collection")

    op.execute("ALTER SEQUENCE collection_id_seq RENAME TO base_dataset_id_seq")
    op.execute("ALTER INDEX collection_pkey RENAME TO base_dataset_pkey")

    op.alter_column("contributed_values", "collection_id", new_column_name="dataset_id")
    op.drop_constraint("contributed_values_collection_id_fkey", "contributed_values", type_="foreignkey")
    op.create_foreign_key(None, "contributed_values", "collection", ["dataset_id"], ["id"], ondelete="cascade")

    # Merge the provenance column with the provenance key in extra
    op.execute(
        sa.text(
            """UPDATE collection
            SET provenance = COALESCE(extra->'provenance', '{}')::jsonb || COALESCE(provenance, '{}')::jsonb"""
        )
    )

    # Update the collection extra, with the removed fields
    op.execute(sa.text("UPDATE collection SET extra = (extra::jsonb - 'provenance')::json"))

    # Remove view columns from base collection
    op.drop_column("collection", "view_url_hdf5")
    op.drop_column("collection", "view_url_plaintext")
    op.drop_column("collection", "view_available")
    op.drop_column("collection", "view_metadata")

    # Finally rename the table
    op.rename_table("collection", "base_dataset")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")
    # ### end Alembic commands ###



"""Add MBE service

Revision ID: f26c30450fdd
Revises: a885e15274b2
Create Date: 2022-05-10 13:32:40.316504

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "f26c30450fdd"
down_revision = "a885e15274b2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "manybody_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(), nullable=False),
        sa.Column("singlepoint_specification_id", sa.Integer(), nullable=False),
        sa.Column("keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.CheckConstraint("program = LOWER(program)", name="ck_manybody_specification_program_lower"),
        sa.ForeignKeyConstraint(
            ["singlepoint_specification_id"],
            ["qc_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("singlepoint_specification_id", "keywords", name="ux_manybody_specification_keys"),
    )
    op.create_index("ix_manybody_specification_keywords", "manybody_specification", ["keywords"], unique=False)
    op.create_index("ix_manybody_specification_program", "manybody_specification", ["program"], unique=False)
    op.create_index(
        "ix_manybody_specification_singlepoint_specification_id",
        "manybody_specification",
        ["singlepoint_specification_id"],
        unique=False,
    )
    op.create_table(
        "manybody_record",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("initial_molecule_id", sa.Integer(), nullable=False),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.Column("results", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.ForeignKeyConstraint(["id"], ["base_record.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["initial_molecule_id"],
            ["molecule.id"],
        ),
        sa.ForeignKeyConstraint(
            ["specification_id"],
            ["manybody_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "manybody_cluster",
        sa.Column("manybody_id", sa.Integer(), nullable=False),
        sa.Column("molecule_id", sa.Integer(), nullable=False),
        sa.Column("fragments", postgresql.ARRAY(sa.Integer()), nullable=False),
        sa.Column("basis", postgresql.ARRAY(sa.Integer()), nullable=False),
        sa.Column("degeneracy", sa.Integer(), nullable=False),
        sa.Column("singlepoint_id", sa.Integer(), nullable=True),
        sa.CheckConstraint("array_length(basis, 1) > 0", name="ck_manybody_cluster_basis"),
        sa.CheckConstraint("array_length(fragments, 1) > 0", name="ck_manybody_cluster_fragments"),
        sa.CheckConstraint("degeneracy > 0", name="ck_manybody_cluster_degeneracy"),
        sa.ForeignKeyConstraint(["manybody_id"], ["manybody_record.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["molecule_id"],
            ["molecule.id"],
        ),
        sa.ForeignKeyConstraint(
            ["singlepoint_id"],
            ["singlepoint_record.id"],
        ),
        sa.PrimaryKeyConstraint("manybody_id", "molecule_id"),
    )
    op.create_index("ix_manybody_cluster_molecule_id", "manybody_cluster", ["molecule_id"], unique=False)
    op.create_index("ix_manybody_cluster_singlepoint_id", "manybody_cluster", ["singlepoint_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_manybody_cluster_singlepoint_id", table_name="manybody_cluster")
    op.drop_index("ix_manybody_cluster_molecule_id", table_name="manybody_cluster")
    op.drop_table("manybody_cluster")
    op.drop_table("manybody_record")
    op.drop_index("ix_manybody_specification_singlepoint_specification_id", table_name="manybody_specification")
    op.drop_index("ix_manybody_specification_program", table_name="manybody_specification")
    op.drop_index("ix_manybody_specification_keywords", table_name="manybody_specification")
    op.drop_table("manybody_specification")
    # ### end Alembic commands ###
